#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from typing import List

from ethtx.models.decoded_model import DecodedTransfer, DecodedBalance, AddressInfo
from .abc import ABISubmoduleAbc

ZERO_ADDRESS = "0x" + 40 * "0"


class ABIBalancesDecoder(ABISubmoduleAbc):
    """Abi Balances Decoder."""

    def decode(self, transfers: List[DecodedTransfer]) -> List:
        """Decode balances."""

        balance_holders = dict()
        balance_tokens = dict()

        for transfer in transfers:
            if transfer.from_address.address != ZERO_ADDRESS:
                balance_holders[
                    transfer.from_address.address
                ] = transfer.from_address.name
            if transfer.to_address.address != ZERO_ADDRESS:
                balance_holders[transfer.to_address.address] = transfer.to_address.name
            balance_tokens[transfer.token_address] = (
                transfer.token_standard,
                transfer.token_symbol,
            )

        balance_sheet: dict = {address: defaultdict(int) for address in balance_holders}

        for transfer in transfers:
            if transfer.from_address.address != ZERO_ADDRESS:
                balance_sheet[transfer.from_address.address][
                    transfer.token_address
                ] -= transfer.value
            if transfer.to_address.address != ZERO_ADDRESS:
                balance_sheet[transfer.to_address.address][
                    transfer.token_address
                ] += transfer.value

        balances = []
        for holder_address in balance_holders:
            tokens = []
            for token_address in balance_sheet[holder_address]:
                if balance_sheet[holder_address][token_address]:
                    token_standard, token_symbol = balance_tokens[token_address]
                    tokens.append(
                        dict(
                            token_address=token_address,
                            token_symbol=token_symbol,
                            token_standard=token_standard,
                            balance=balance_sheet[holder_address][token_address],
                        )
                    )
            if tokens:
                holder_name = balance_holders[holder_address]
                balances.append(
                    DecodedBalance(
                        holder=AddressInfo(holder_address, holder_name), tokens=tokens
                    )
                )

        return balances



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import (
    EventSemantics,
    ParameterSemantics,
    TransformationSemantics,
)

lognote_event_v1 = EventSemantics(
    "0xd3ff30f94bb4ebb4f3d773ea26b6efc7328b9766f99f19dff6f01392138be46d",
    False,
    "LogNote",
    [
        ParameterSemantics("sig", "bytes4", [], True),
        ParameterSemantics("arg1", "bytes32", [], True),
        ParameterSemantics("arg2", "bytes32", [], True),
        ParameterSemantics("arg3", "bytes32", [], True),
        ParameterSemantics("data", "bytes", [], False, True),
    ],
)

lognote_transformation_v1 = {
    "sig": TransformationSemantics(transformed_type="ignore"),
    "arg1": TransformationSemantics(transformed_type="ignore"),
    "arg2": TransformationSemantics(transformed_type="ignore"),
    "arg3": TransformationSemantics(transformed_type="ignore"),
    "data": TransformationSemantics(
        transformed_type="call", transformation="decode_call(__contract__, data)"
    ),
}

lognote_event_v2 = EventSemantics(
    "0xd3d8bec38a91a5f4411247483bc030a174e77cda9c0351924c759f41453aa5e8",
    False,
    "LogNote",
    [
        ParameterSemantics("sig", "bytes4", [], True),
        ParameterSemantics("user", "address", [], True),
        ParameterSemantics("arg1", "bytes32", [], True),
        ParameterSemantics("arg2", "bytes32", [], True),
        ParameterSemantics("data", "bytes", [], False, True),
    ],
)

lognote_transformation_v2 = {
    "sig": TransformationSemantics(transformed_type="ignore"),
    "arg1": TransformationSemantics(transformed_type="ignore"),
    "arg2": TransformationSemantics(transformed_type="ignore"),
    "data": TransformationSemantics(
        transformed_type="call", transformation="decode_call(__contract__, data)"
    ),
}

lognote_event_v3 = EventSemantics(
    "0x644843f351d3fba4abcd60109eaff9f54bac8fb8ccf0bab941009c21df21cf31",
    False,
    "LogNote",
    [
        ParameterSemantics("sig", "bytes4", [], True),
        ParameterSemantics("guy", "address", [], True),
        ParameterSemantics("foo", "bytes32", [], True),
        ParameterSemantics("bar", "bytes32", [], True),
        ParameterSemantics("wad", "uint256", [], False),
        ParameterSemantics("fax", "bytes", [], False, True),
    ],
)

lognote_transformation_v3 = {
    "sig": TransformationSemantics(transformed_type="ignore"),
    "foo": TransformationSemantics(transformed_type="ignore"),
    "bar": TransformationSemantics(transformed_type="ignore"),
    "wad": TransformationSemantics(transformation="wad / 10**18"),
    "fax": TransformationSemantics(
        transformed_type="call", transformation="decode_call(__contract__, fax)"
    ),
}

logcall_event = EventSemantics(
    "0x25fce1fe01d9b241fda40b2152ddd6f4ba063fcfb3c2c81dddf84ee20d3f341f",
    False,
    "LOG_CALL",
    [
        ParameterSemantics("sig", "bytes4", [], True),
        ParameterSemantics("caller", "address", [], True),
        ParameterSemantics("data", "bytes", [], False, True),
    ],
)

logcall_transformation = {
    "sig": TransformationSemantics(transformed_type="ignore"),
    "data": TransformationSemantics(
        transformed_type="call", transformation="decode_call(__contract__, data)"
    ),
}

anonymous_events = {
    lognote_event_v1.signature: lognote_transformation_v1,
    lognote_event_v2.signature: lognote_transformation_v2,
    lognote_event_v3.signature: lognote_transformation_v3,
    logcall_event.signature: logcall_transformation,
}



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
from functools import lru_cache
from typing import List, Dict, Optional

from web3 import Web3
from web3.datastructures import AttributeDict
from web3.types import BlockData, TxData, TxReceipt, HexStr

from ..exceptions import Web3ConnectionException, ProcessingException
from ..models.objects_model import (
    Block,
    Transaction,
    BlockMetadata,
    TransactionMetadata,
    Call,
)
from ..models.w3_model import W3Block, W3Transaction, W3Receipt, W3CallTree, W3Log
from ..semantics.standards import erc20

log = logging.getLogger(__name__)


def connect_chain(
    http_hook: str = None, ipc_hook: str = None, ws_hook: str = None
) -> Web3 or None:
    if http_hook:
        method = "HTTP"
        provider = Web3.HTTPProvider
        hook = http_hook
    elif ipc_hook:
        method = "IPC"
        provider = Web3.IPCProvider
        hook = ipc_hook
    elif ws_hook:
        method = "Websocket"
        provider = Web3.WebsocketProvider
        hook = ws_hook
    else:
        method = "IPC"
        provider = Web3.IPCProvider
        hook = "\\\\.\\pipe\\geth.ipc"

    try:
        w3 = Web3(provider(hook, request_kwargs={"timeout": 600}))
        if w3.isConnected():
            log.info(
                "Connected to %s: %s with latest block %s.",
                method,
                hook,
                w3.eth.block_number,
            )
            return w3
        else:
            log.info("%s connection to %s failed.", method, hook)
            raise Web3ConnectionException()
    except Exception as exc:
        log.warning("Node connection %s: %s failed.", method, hook, exc_info=exc)
        raise


class NodeDataProvider:

    default_chain: str

    def __init__(self, default_chain=None):
        self.default_chain = default_chain

    def get_block(
        self, block_number: int, chain_id: Optional[str] = None
    ) -> BlockMetadata:
        ...

    def get_transaction(
        self, tx_hash: str, chain_id: Optional[str] = None
    ) -> TransactionMetadata:
        ...

    def get_full_transaction(
        self, tx_hash: str, chain_id: Optional[str] = None
    ) -> Transaction:
        ...

    def get_calls(self, tx_hash: str, chain_id: Optional[str] = None) -> Call:
        ...


class Web3Provider(NodeDataProvider):
    chain: Web3

    def __init__(self, nodes: Dict[str, str], default_chain=None):
        super().__init__(default_chain)
        self.nodes = nodes

    def _get_node_connection(self, chain_id: Optional[str] = None) -> Web3:
        chain_id = chain_id or self.default_chain

        if chain_id is None:
            raise ProcessingException(
                "chain_id must be provided as an argument or constructor default"
            )
        return connect_chain(http_hook=self.nodes[chain_id])

    # get the raw block data from the node
    @lru_cache(maxsize=512)
    def get_block(self, block_number: int, chain_id: Optional[str] = None) -> W3Block:
        chain = self._get_node_connection(chain_id)
        raw_block: BlockData = chain.eth.get_block(block_number)
        block = W3Block(
            chain_id=chain_id or self.default_chain,
            difficulty=raw_block.difficulty,
            extraData=raw_block.extraData,
            gasLimit=raw_block.gasLimit,
            gasUsed=raw_block.gasUsed,
            hash=raw_block.hash,
            logsBloom=raw_block.logsBloom,
            miner=raw_block.miner,
            nonce=raw_block.get("nonce", 0),
            number=raw_block.number,
            parentHash=raw_block.parentHash,
            receiptsRoot=raw_block.receiptsRoot,
            sha3Uncles=raw_block.sha3Uncles,
            size=raw_block.size,
            stateRoot=raw_block.stateRoot,
            timestamp=raw_block.timestamp,
            totalDifficulty=raw_block.totalDifficulty,
            transactions=raw_block.transactions,
            transactionsRoot=raw_block.transactionsRoot,
            uncles=raw_block.uncles,
        )

        return block

    # get the raw transaction data from the node
    @lru_cache(maxsize=512)
    def get_transaction(
        self, tx_hash: str, chain_id: Optional[str] = None
    ) -> W3Transaction:
        chain = self._get_node_connection(chain_id)
        raw_tx: TxData = chain.eth.get_transaction(HexStr(tx_hash))
        transaction = W3Transaction(
            chain_id=chain_id or self.default_chain,
            blockHash=raw_tx.blockHash,
            blockNumber=raw_tx.blockNumber,
            from_address=raw_tx["from"],
            gas=raw_tx.gas,
            gasPrice=raw_tx.gasPrice,
            hash=raw_tx.hash,
            input=raw_tx.input,
            nonce=raw_tx.nonce,
            r=raw_tx.r,
            s=raw_tx.s,
            to=raw_tx.to,
            transactionIndex=raw_tx.transactionIndex,
            v=raw_tx.v,
            value=raw_tx.value,
        )

        return transaction

    @lru_cache(maxsize=512)
    def get_receipt(self, tx_hash: str, chain_id: Optional[str] = None) -> W3Receipt:
        chain = self._get_node_connection(chain_id)
        raw_receipt: TxReceipt = chain.eth.get_transaction_receipt(tx_hash)
        _root = raw_receipt.root if hasattr(raw_receipt, "root") else None

        _logs = [
            W3Log(
                tx_hash=tx_hash,
                chain_id=chain_id or self.default_chain,
                address=_log.address,
                blockHash=_log.blockHash,
                blockNumber=_log.blockNumber,
                data=_log.data,
                logIndex=_log.logIndex,
                removed=_log.removed,
                topics=_log.topics,
                transactionHash=_log.transactionHash,
                transactionIndex=_log.transactionIndex,
            )
            for _log in raw_receipt.logs
        ]

        receipt = W3Receipt(
            tx_hash=tx_hash,
            chain_id=chain_id or self.default_chain,
            blockHash=raw_receipt.blockHash,
            blockNumber=raw_receipt.blockNumber,
            contractAddress=raw_receipt.contractAddress,
            cumulativeGasUsed=raw_receipt.cumulativeGasUsed,
            from_address=raw_receipt["from"],
            gasUsed=raw_receipt.gasUsed,
            logs=_logs,
            logsBloom=raw_receipt.logsBloom,
            root=_root,
            status=raw_receipt.status,
            to_address=raw_receipt.to,
            transactionHash=raw_receipt.transactionHash,
            transactionIndex=raw_receipt.transactionIndex,
        )

        return receipt

    @staticmethod
    def _get_custom_calls_tracer():
        return open(os.path.join(os.path.dirname(__file__), "static/tracer.js")).read()

    @lru_cache(maxsize=512)
    def get_calls(self, tx_hash: str, chain_id: Optional[str] = None) -> W3CallTree:
        # tracer is a temporary fixed implementation of geth tracer
        chain = self._get_node_connection(chain_id)
        tracer = self._get_custom_calls_tracer()
        response = chain.manager.request_blocking(
            "debug_traceTransaction", [tx_hash, {"tracer": tracer}]
        )

        return self._create_call_from_debug_trace_tx(
            tx_hash, chain_id or self.default_chain, response
        )

    # get the contract bytecode hash from the node
    @lru_cache(maxsize=512)
    def get_code_hash(
        self, contract_address: str, chain_id: Optional[str] = None
    ) -> str:
        chain = self._get_node_connection(chain_id)
        byte_code = chain.eth.get_code(Web3.toChecksumAddress(contract_address))
        code_hash = Web3.keccak(byte_code).hex()
        return code_hash

    # get the erc20 token data from the node
    @lru_cache(maxsize=512)
    def get_erc20_token(
        self,
        token_address: str,
        contract_name: str,
        functions,
        chain_id: Optional[str] = None,
    ):

        name_abi = symbol_abi = decimals_abi = ""

        if functions:
            for function in functions.values():
                if (
                    function.name == "name"
                    and len(function.inputs) == 0
                    and len(function.outputs) == 1
                ):
                    name_type = function.outputs[0].parameter_type
                    name_abi = (
                        '{"name":"name", "constant":true, "payable":false, "type":"function", '
                        ' "inputs":[], "outputs":[{"name":"","type":"%s"}]}' % name_type
                    )

                elif (
                    function.name == "symbol"
                    and len(function.inputs) == 0
                    and len(function.outputs) == 1
                ):
                    symbol_type = function.outputs[0].parameter_type
                    symbol_abi = (
                        '{"name":"symbol", "constant":true, "payable":false,"type":"function", '
                        ' "inputs":[], "outputs":[{"name":"","type":"%s"}]}'
                        % symbol_type
                    )

                elif (
                    function.name in ["decimals", "dec"]
                    and len(function.inputs) == 0
                    and len(function.outputs) == 1
                ):
                    decimals_type = function.outputs[0].parameter_type
                    decimals_abi = (
                        '{"name":"decimals", "constant":true, "payable":false,"type":"function", '
                        ' "inputs":[], "outputs":[{"name":"","type":"%s"}]}'
                        % decimals_type
                    )

        abi = f'[{",".join([name_abi, symbol_abi, decimals_abi])}]'

        try:
            chain = self._get_node_connection(chain_id)
            token = chain.eth.contract(
                address=Web3.toChecksumAddress(token_address), abi=abi
            )
            name = token.functions.name().call() if name_abi else contract_name
            if type(name) == bytes:
                name = name.decode("utf-8").replace("\x00", "")

            symbol = token.functions.symbol().call() if symbol_abi else contract_name
            if type(symbol) == bytes:
                symbol = symbol.decode("utf-8").replace("\x00", "")

            decimals = token.functions.decimals().call() if decimals_abi else 18
        except Exception:
            name = symbol = contract_name
            decimals = 18

        return dict(address=token_address, symbol=symbol, name=name, decimals=decimals)

    # guess if the contract is and erc20 token and get the data
    @lru_cache(maxsize=512)
    def guess_erc20_token(self, contract_address, chain_id: Optional[str] = None):
        chain = self._get_node_connection(chain_id)

        byte_code = chain.eth.get_code(Web3.toChecksumAddress(contract_address)).hex()

        if all(
            "63" + signature[2:] in byte_code
            for signature in (
                erc20.erc20_transfer_function.signature,
                erc20.erc20_transferFrom_function.signature,
                erc20.erc20_approve_function.signature,
            )
        ) and all(
            signature[2:] in byte_code
            for signature in (
                erc20.erc20_transfer_event.signature,
                erc20.erc20_approval_event.signature,
            )
        ):

            name_abi = (
                '{"name":"name", "constant":true, "payable":false,'
                ' "type":"function", "inputs":[], "outputs":[{"name":"","type":"string"}]}'
            )
            symbol_abi = (
                '{"name":"symbol", "constant":true, "payable":false,'
                '"type":"function", "inputs":[], "outputs":[{"name":"","type":"string"}]}'
            )
            decimals_abi = (
                '{"name":"decimals", "constant":true, "payable":false,'
                '"type":"function",  "inputs":[], "outputs":[{"name":"","type":"uint8"}]}'
            )

            abi = f'[{",".join([name_abi, symbol_abi, decimals_abi])}]'

            try:
                token = chain.eth.contract(
                    address=Web3.toChecksumAddress(contract_address), abi=abi
                )
                name = token.functions.name().call()
                symbol = token.functions.symbol().call()
                decimals = token.functions.decimals().call()

                return dict(
                    address=contract_address,
                    symbol=symbol,
                    name=name,
                    decimals=decimals,
                )

            except Exception:
                pass

        return None

    # guess if the contract is and erc20 token proxy and get the data
    @lru_cache(maxsize=512)
    def guess_erc20_proxy(self, contract_address, chain_id: Optional[str] = None):
        chain = self._get_node_connection(chain_id)

        name_abi = (
            '{"name":"name", "constant":true, "payable":false,'
            ' "type":"function", "inputs":[], "outputs":[{"name":"","type":"string"}]}'
        )
        symbol_abi = (
            '{"name":"symbol", "constant":true, "payable":false,'
            '"type":"function", "inputs":[], "outputs":[{"name":"","type":"string"}]}'
        )
        decimals_abi = (
            '{"name":"decimals", "constant":true, "payable":false,'
            '"type":"function",  "inputs":[], "outputs":[{"name":"","type":"uint8"}]}'
        )

        abi = f'[{",".join([name_abi, symbol_abi, decimals_abi])}]'

        try:
            token = chain.eth.contract(
                address=Web3.toChecksumAddress(contract_address), abi=abi
            )
            name = token.functions.name().call()
            symbol = token.functions.symbol().call()
            decimals = token.functions.decimals().call()

            return dict(symbol=symbol, name=name, decimals=decimals)

        except Exception:
            pass

        return None

    @staticmethod
    def _create_call_from_debug_trace_tx(
        tx_hash: str, chain_id: str, input_rpc: AttributeDict
    ) -> W3CallTree:
        def prep_raw_dict(dct: [AttributeDict, Dict]):
            if not isinstance(dct, dict):
                dct = dct.__dict__
            dct["from_address"] = dct.pop("from")
            dct["to_address"] = dct.pop("to")
            dct["input"] = dct.pop("input", "0x")
            dct["output"] = dct.pop("output", "0x")
            calls = dct.pop("calls", [])
            return dct, calls

        obj = input_rpc.__dict__
        tmp_call_tree = []

        w3input, main_parent_calls = prep_raw_dict(obj)
        main_parent = W3CallTree(tx_hash, chain_id, **w3input)
        for main_parent_call in main_parent_calls:
            w3input, main_parent_calls = prep_raw_dict(main_parent_call)
            main_parent_child = W3CallTree(tx_hash, chain_id, **w3input)
            main_parent.calls.append(main_parent_child)
            if len(main_parent_calls) > 0:
                tmp_call_tree.append(
                    {"parent": main_parent_child, "children": main_parent_calls}
                )

        while len(tmp_call_tree) != 0:
            new_call_tree = []

            for pair in tmp_call_tree:

                parent_call: W3CallTree = pair["parent"]
                child_calls: List = pair["children"]

                if child_calls is not None:
                    for child_call in child_calls:
                        w3input, child_child_call = prep_raw_dict(child_call)
                        child = W3CallTree(tx_hash, chain_id, **w3input)
                        parent_call.calls.append(child)

                        if len(child_call) > 0:
                            new_call_tree.append(
                                {"parent": child, "children": child_child_call}
                            )

            tmp_call_tree = new_call_tree

        return main_parent

    @lru_cache(maxsize=512)
    def get_full_block(self, block_number: int, chain_id: Optional[str] = None):

        w3block = self.get_block(block_number, chain_id)
        w3transactions = [
            (
                self.get_transaction(tx_hash.hex(), chain_id),
                self.get_receipt(tx_hash.hex(), chain_id),
                self.get_calls(tx_hash.hex(), chain_id),
            )
            for tx_hash in w3block.transactions
        ]

        return Block.from_raw(
            chain_id=chain_id or self.default_chain,
            w3block=w3block,
            w3transactions=w3transactions,
        )

    @lru_cache(maxsize=512)
    def get_full_transaction(self, tx_hash: str, chain_id: Optional[str] = None):

        w3transaction = self.get_transaction(tx_hash, chain_id)
        w3receipt = self.get_receipt(tx_hash, chain_id)
        w3calltree = self.get_calls(tx_hash, chain_id)

        return Transaction.from_raw(
            w3transaction=w3transaction, w3receipt=w3receipt, w3calltree=w3calltree
        )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from datetime import datetime
from typing import List, Any, Optional

from ethtx.models.objects_model import BlockMetadata, TransactionMetadata
from ethtx.utils.pickable import JsonObject


class AddressInfo(JsonObject):
    address: str
    name: str
    badge: Optional[str]

    def __init__(self, address: str, name: str, badge: Optional[str] = ""):
        self.address = address
        self.name = name
        self.badge = badge

    def __eq__(self, other):
        if isinstance(other, AddressInfo):
            return (
                self.address == other.address
                and self.name == other.name
                and self.badge == other.badge
            )
        return False


class DecodedTransactionMetadata(JsonObject):
    chain_id: str
    tx_hash: str
    block_number: int
    timestamp: datetime
    gas_price: int
    sender: AddressInfo
    receiver: AddressInfo
    tx_index: int
    tx_value: int
    eth_price: float
    gas_limit: int
    gas_used: int
    success: bool

    def __init__(
        self,
        chain_id: str,
        tx_hash: str,
        block_number: int,
        timestamp: datetime,
        gas_price: int,
        sender: AddressInfo,
        receiver: AddressInfo,
        tx_index: int,
        tx_value: int,
        eth_price: float,
        gas_limit: int,
        gas_used: int,
        success: bool,
    ):
        self.chain_id = chain_id
        self.tx_hash = tx_hash
        self.block_number = block_number
        self.timestamp = timestamp
        self.gas_price = gas_price
        self.sender = sender
        self.receiver = receiver
        self.tx_index = tx_index
        self.tx_value = tx_value
        self.eth_price = eth_price
        self.gas_limit = gas_limit
        self.gas_used = gas_used
        self.success = success


class Argument(JsonObject):
    name: str
    type: str
    value: Any

    def __init__(self, name: str, type: str, value: Any):
        self.name = name
        self.type = type
        self.value = value

    def __eq__(self, other):
        if isinstance(other, Argument):
            return (
                self.name == other.name
                and self.type == other.type
                and self.value == other.value
            )
        return False


class DecodedEvent(JsonObject):
    chain_id: str
    tx_hash: str
    contract: AddressInfo
    index: int
    event_signature: str
    event_name: str
    parameters: List[Argument]

    def __init__(
        self,
        chain_id: str,
        tx_hash: str,
        contract_address: str,
        contract_name: str,
        index: int,
        event_signature: str,
        event_name: str,
        parameters: List[Argument],
    ):
        self.chain_id = chain_id
        self.tx_hash = tx_hash
        self.contract = AddressInfo(contract_address, contract_name)
        self.contract_name = contract_name
        self.index = index
        self.event_signature = event_signature
        self.event_name = event_name
        self.parameters = parameters

    def __eq__(self, other):
        if isinstance(other, DecodedEvent):
            return (
                self.chain_id == other.chain_id
                and self.tx_hash == other.tx_hash
                and self.contract == other.contract
                and self.contract_name == other.contract_name
                and self.index == other.index
                and self.event_signature == other.event_signature
                and self.event_name == other.event_name
                and self.parameters == other.parameters
            )
        return False


class DecodedCall(JsonObject):
    chain_id: str
    tx_hash: str
    call_id: str
    call_type: str
    from_address: AddressInfo
    to_address: AddressInfo
    value: int
    function_signature: str
    function_name: str
    arguments: List[Argument]
    outputs: List[Argument]
    gas_used: int
    error: str
    status: bool
    subcalls: List[DecodedCall]

    def __init__(
        self,
        chain_id: str,
        tx_hash: str,
        call_id: str,
        call_type: str,
        from_address: str,
        from_name: str,
        to_address: str,
        to_name: str,
        value: int,
        function_signature: str,
        function_name: str,
        arguments: List[Argument],
        outputs: List[Argument],
        gas_used: int,
        error: str,
        status: bool,
        indent: int,
        subcalls: Optional[List[DecodedCall]] = None,
    ):
        self.chain_id = chain_id
        self.tx_hash = tx_hash
        self.call_id = call_id
        self.call_type = call_type
        self.from_address = AddressInfo(from_address, from_name)
        self.to_address = AddressInfo(to_address, to_name)
        self.to_name = to_name
        self.value = value
        self.function_signature = function_signature
        self.function_name = function_name
        self.arguments = arguments
        self.outputs = outputs
        self.gas_used = gas_used
        self.error = error
        self.status = status
        self.indent = indent
        self.subcalls = subcalls if subcalls else []

    def __eq__(self, other):
        if isinstance(other, DecodedCall):
            return (
                self.chain_id == other.chain_id
                and self.tx_hash == other.tx_hash
                and self.call_type == other.call_type
                and self.from_address == other.from_address
                and self.to_address == other.to_address
                and self.to_name == other.to_name
                and self.value == other.value
                and self.function_signature == other.function_signature
                and self.function_name == other.function_name
                and self.arguments == other.arguments
                and self.outputs == other.outputs
                and self.gas_used == other.gas_used
                and self.error == other.error
                and self.status == other.status
                and self.subcalls == other.subcalls
            )

        return False


class DecodedTransfer(JsonObject):
    from_address: AddressInfo
    to_address: AddressInfo
    token_address: Optional[str]
    token_symbol: str
    token_standard: Optional[str]
    value: float

    def __init__(
        self,
        from_address: AddressInfo,
        to_address: AddressInfo,
        token_standard: Optional[str],
        token_address: Optional[str],
        token_symbol: str,
        value: float,
    ):
        self.from_address = from_address
        self.to_address = to_address
        self.token_address = token_address
        self.token_symbol = token_symbol
        self.token_standard = token_standard
        self.value = value


class DecodedBalance(JsonObject):
    holder: AddressInfo
    tokens: List[dict]

    def __init__(self, holder: AddressInfo, tokens: List[dict]):
        self.holder = holder
        self.tokens = tokens


class DecodedTransaction(JsonObject):
    block_metadata: BlockMetadata
    metadata: TransactionMetadata
    events: List[DecodedEvent]
    calls: Optional[DecodedCall]
    transfers: List[DecodedTransfer]
    balances: List[DecodedBalance]
    status: bool

    def __init__(
        self,
        block_metadata: BlockMetadata,
        tx_metadata: TransactionMetadata,
        events: List[DecodedEvent],
        calls: Optional[DecodedCall],
        transfers: List[DecodedTransfer],
        balances: List[DecodedBalance],
    ):
        self.block_metadata = block_metadata
        self.metadata = tx_metadata
        self.events = events
        self.calls = calls
        self.transfers = transfers
        self.balances = balances
        self.status = False



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from typing import Dict, Optional

from pymongo.database import Database as MongoDatabase


class ISemanticsDatabase(ABC):
    """Semantics Database. Represents raw interface required to be
    implemented by a database that provides persistent
    data about semantics"""

    def get_address_semantics(self, chain_id: str, address: str) -> Optional[Dict]:
        ...

    def get_contract_semantics(self, code_hash: str) -> Optional[Dict]:
        ...

    def get_signature_semantics(self, signature_hash: str) -> Optional[Dict]:
        ...

    def insert_contract(self, contract: dict, update_if_exist: bool = False):
        ...

    def insert_address(self, address_data: dict, update_if_exist: bool = False):
        ...

    def insert_signature(self, signature, update_if_exist: bool = False):
        ...


class MongoCollections:
    ADDRESSES = "addresses"
    CONTRACTS = "contracts"
    SIGNATURES = "signatures"


class MongoSemanticsDatabase(ISemanticsDatabase):
    def get_collection_count(self):
        return len(self._db.list_collection_names())

    def __init__(self, db: MongoDatabase):
        self._db = db
        self._addresses = self._db["addresses"]
        self._contracts = self._db["contracts"]
        self._signatures = self._db["signatures"]

    def get_address_semantics(self, chain_id, address) -> Optional[Dict]:
        _id = f"{chain_id}-{address}"
        return self._addresses.find_one({"_id": _id}, {"_id": 0})

    def get_signature_semantics(self, signature_hash):
        return self._signatures.find_one({"_id": signature_hash}, {"_id": 0})

    def get_contract_semantics(self, code_hash):
        """Contract hashes are always the same, no mather what chain we use, so there is no need
        to use chain_id"""
        return self._contracts.find_one({"_id": code_hash}, {"_id": 0})

    def insert_contract(self, contract, update_if_exist=False):
        contract_with_id = {"_id": contract["code_hash"], **contract}

        if update_if_exist:
            self._contracts.replace_one(
                {"_id": contract_with_id["_id"]}, contract_with_id, upsert=True
            )
        else:
            self._contracts.insert_one(contract_with_id)

    def insert_address(self, address, update_if_exist=False):
        address_with_id = {
            "_id": f"{address['chain_id']}-{address['address']}",
            **address,
        }

        if update_if_exist:
            self._addresses.replace_one(
                {"_id": address_with_id["_id"]}, address_with_id, upsert=True
            )
        else:
            self._addresses.insert_one(address_with_id)

    def insert_signature(self, signature, update_if_exist=False):
        signature_with_id = {"_id": signature["hash"], **signature}

        if update_if_exist:
            self._signatures.replace_one(
                {"_id": signature_with_id["_id"]}, signature_with_id, upsert=True
            )
        else:
            self._signatures.insert_one(signature_with_id)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

import jsonpickle


class JsonObject:
    """Model to json - base class."""

    def json_str(self) -> str:
        """Return object as json string."""
        return jsonpickle.encode(self, make_refs=False, unpicklable=False)

    def json(self) -> Dict:
        """Return object as json."""
        return jsonpickle.decode(self.json_str())



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Union, Optional

from ethtx.models.decoded_model import DecodedEvent
from ethtx.models.objects_model import TransactionMetadata, Event
from ethtx.semantics.standards.erc20 import ERC20_EVENTS
from .abc import ABISubmoduleAbc
from ..decoders.parameters import decode_event_parameters


class ABIEventsDecoder(ABISubmoduleAbc):
    """ABI Events Decoder."""

    def decode(
        self,
        events: Union[Event, List[Event]],
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
        chain_id: Optional[str] = None,
    ) -> Union[DecodedEvent, List[DecodedEvent]]:
        """Return list of decoded events."""
        if isinstance(events, list):
            return (
                [
                    self.decode_event(
                        event, transaction, delegations, token_proxies, chain_id
                    )
                    for event in events
                ]
                if events
                else []
            )

        return self.decode_event(
            events, transaction, delegations, token_proxies, chain_id
        )

    def decode_event(
        self,
        event: Event,
        transaction: TransactionMetadata,
        delegations: Dict[str, set] = None,
        token_proxies: Dict[str, dict] = None,
        chain_id: str = None,
    ) -> DecodedEvent:

        if event.topics:
            event_signature = event.topics[0]
        else:
            event_signature = None

        anonymous = False
        chain_id = chain_id or self._default_chain

        event_abi = self._repository.get_event_abi(
            chain_id, event.contract, event_signature
        )

        if not event_abi:

            if event_signature in ERC20_EVENTS:
                # try standard ERC20 events
                event_abi = ERC20_EVENTS[event_signature]

            if not event_abi:
                # if signature is not known but there is exactly one anonymous event in tha ABI
                # we can assume that this is this the anonymous one (e.g. Maker's LogNote)
                event_abi = self._repository.get_anonymous_event_abi(
                    chain_id, event.contract
                )
                if event_abi:
                    anonymous = True

            if not event_abi and event.contract in delegations:
                # try to find signature in delegate-called contracts
                for delegate in delegations[event.contract]:
                    event_abi = self._repository.get_event_abi(
                        chain_id, delegate, event_signature
                    )
                    if event_abi:
                        break

        contract_name = self._repository.get_address_label(
            chain_id, event.contract, token_proxies
        )
        event_name = event_abi.name if event_abi else event_signature
        parameters = decode_event_parameters(
            event.log_data, event.topics, event_abi, anonymous
        )

        return DecodedEvent(
            chain_id=chain_id,
            tx_hash=transaction.tx_hash,
            contract_address=event.contract,
            contract_name=contract_name,
            index=event.log_index,
            event_signature=event_signature,
            event_name=event_name,
            parameters=parameters,
        )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import lru_cache
from typing import Optional, List

from ethtx.decoders.decoders.semantics import decode_events_and_functions
from ethtx.models.semantics_model import (
    AddressSemantics,
    ContractSemantics,
    ParameterSemantics,
    ERC20Semantics,
    TransformationSemantics,
    FunctionSemantics,
    EventSemantics,
)
from ethtx.providers.etherscan_provider import EtherscanProvider
from ethtx.providers.semantic_providers.semantics_database import ISemanticsDatabase
from ethtx.providers.web3_provider import Web3Provider
from ethtx.semantics.protocols_router import amend_contract_semantics
from ethtx.semantics.standards.erc20 import ERC20_FUNCTIONS, ERC20_EVENTS
from ethtx.semantics.standards.erc721 import ERC721_FUNCTIONS, ERC721_EVENTS


class SemanticsRepository:
    def __init__(
        self,
        database_connection: ISemanticsDatabase,
        etherscan_provider: EtherscanProvider,
        web3provider: Web3Provider,
    ):
        self.database = database_connection
        self.etherscan = etherscan_provider
        self._web3provider = web3provider
        self._records: Optional[List] = None

    def record(self):
        """Records is an array used to hold semantics used in tx decing process.
        This recording is used just for logging"""
        self._records = []

    def end_record(self) -> List:
        tmp_records = self._records
        self._records = None
        return tmp_records

    def get_raw_semantics(self, address, chain_id=None):
        def decode_parameter(_parameter):
            components_semantics = []
            for component in _parameter["components"]:
                components_semantics.append(decode_parameter(component))

            decoded_parameter = ParameterSemantics(
                _parameter["parameter_name"],
                _parameter["parameter_type"],
                components_semantics,
                _parameter["indexed"],
                _parameter["dynamic"],
            )

            return decoded_parameter

        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        raw_address_semantics = self.database.get_address_semantics(chain_id, address)

        if raw_address_semantics:

            if raw_address_semantics.get("erc20"):
                erc20_semantics = ERC20Semantics(
                    raw_address_semantics["erc20"]["name"],
                    raw_address_semantics["erc20"]["symbol"],
                    raw_address_semantics["erc20"]["decimals"],
                )
            else:
                erc20_semantics = None

            if raw_address_semantics["contract"] == ZERO_HASH:
                contract_semantics = ContractSemantics(
                    raw_address_semantics["contract"], "EOA", dict(), dict(), dict()
                )

            else:

                raw_contract_semantics = self.database.get_contract_semantics(
                    raw_address_semantics["contract"]
                )
                events = dict()

                for signature, event in raw_contract_semantics["events"].items():

                    parameters_semantics = []
                    for parameter in event["parameters"]:
                        parameters_semantics.append(decode_parameter(parameter))

                    events[signature] = EventSemantics(
                        signature,
                        event["anonymous"],
                        event["name"],
                        parameters_semantics,
                    )

                functions = dict()
                for signature, function in raw_contract_semantics["functions"].items():

                    inputs_semantics = []
                    for parameter in function["inputs"]:
                        inputs_semantics.append(decode_parameter(parameter))
                    outputs_semantics = []
                    for parameter in function["outputs"]:
                        outputs_semantics.append(decode_parameter(parameter))

                    functions[signature] = FunctionSemantics(
                        signature, function["name"], inputs_semantics, outputs_semantics
                    )

                transformations = dict()
                for signature, parameters_transformations in raw_contract_semantics[
                    "transformations"
                ].items():
                    transformations[signature] = dict()
                    for parameter, transformation in parameters_transformations.items():
                        transformations[signature][parameter] = TransformationSemantics(
                            transformation["transformed_name"],
                            transformation["transformed_type"],
                            transformation["transformation"],
                        )

                contract_semantics = ContractSemantics(
                    raw_contract_semantics["code_hash"],
                    raw_contract_semantics["name"],
                    events,
                    functions,
                    transformations,
                )

            address_semantics = AddressSemantics(
                chain_id,
                address,
                raw_address_semantics["name"],
                raw_address_semantics["is_contract"],
                contract_semantics,
                raw_address_semantics["standard"],
                erc20_semantics,
            )

            return address_semantics

        else:
            return None

    @lru_cache(maxsize=128)
    def _get_semantics(self, chain_id: str, address: str) -> Optional[AddressSemantics]:

        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        address_semantics = self.get_raw_semantics(address, chain_id)
        if not address_semantics:
            # try to read the semantics form the Etherscan provider
            provider = self._web3provider
            code_hash = provider.get_code_hash(address)

            if code_hash != ZERO_HASH:
                # smart contract
                raw_semantics, decoded = self.etherscan.get_contract_abi(
                    chain_id, address
                )
                if decoded and raw_semantics:
                    # raw semantics received from Etherscan
                    events, functions = decode_events_and_functions(
                        raw_semantics["abi"]
                    )
                    standard, standard_semantics = self._decode_standard_semantics(
                        address, raw_semantics["name"], events, functions
                    )
                    if standard == "ERC20":
                        erc20_semantics = standard_semantics
                    else:
                        proxy_erc20 = provider.guess_erc20_proxy(address)
                        if proxy_erc20:
                            erc20_semantics = ERC20Semantics(**proxy_erc20)
                        else:
                            erc20_semantics = None
                    contract_semantics = ContractSemantics(
                        code_hash, raw_semantics["name"], events, functions, dict()
                    )
                    address_semantics = AddressSemantics(
                        chain_id,
                        address,
                        raw_semantics["name"],
                        True,
                        contract_semantics,
                        standard,
                        erc20_semantics,
                    )

                else:
                    # try to guess if the address is a toke
                    potential_erc20_semantics = provider.guess_erc20_token(address)
                    if potential_erc20_semantics:
                        standard = "ERC20"
                        erc20_semantics = ERC20Semantics(
                            potential_erc20_semantics["name"],
                            potential_erc20_semantics["symbol"],
                            potential_erc20_semantics["decimals"],
                        )
                    else:
                        standard = None
                        erc20_semantics = None

                    contract_semantics = ContractSemantics(
                        code_hash, address, dict(), dict(), dict()
                    )
                    address_semantics = AddressSemantics(
                        chain_id,
                        address,
                        address,
                        True,
                        contract_semantics,
                        standard,
                        erc20_semantics,
                    )

            else:
                # externally owned address
                contract_semantics = ContractSemantics(
                    ZERO_HASH, "EOA", dict(), dict(), dict()
                )
                address_semantics = AddressSemantics(
                    chain_id, address, address, False, contract_semantics, None, None
                )

            self.update_semantics(address_semantics)

        # amend semantics with locally stored updates
        amend_contract_semantics(address_semantics.contract)

        return address_semantics

    def _decode_standard_semantics(self, address, name, events, functions):
        standard = None
        standard_semantics = None

        if all(erc20_event in events for erc20_event in ERC20_EVENTS) and all(
            erc20_function in functions for erc20_function in ERC20_FUNCTIONS
        ):
            standard = "ERC20"
            try:
                provider = self._web3provider
                token_data = provider.get_erc20_token(address, name, functions)
                standard_semantics = ERC20Semantics(
                    token_data["name"], token_data["symbol"], token_data["decimals"]
                )
            except Exception:
                standard_semantics = ERC20Semantics(name, name, 18)
        elif all(erc721_event in events for erc721_event in ERC721_EVENTS) and all(
            erc721_function in functions for erc721_function in ERC721_FUNCTIONS
        ):
            standard = "ERC721"
            standard_semantics = None

        return standard, standard_semantics

    @lru_cache(maxsize=128)
    def get_event_abi(self, chain_id, address, signature):
        semantics = self._get_semantics(chain_id, address)
        event_semantics = (
            semantics.contract.events.get(signature) if semantics else None
        )

        return event_semantics

    @lru_cache(maxsize=128)
    def get_transformations(self, chain_id, address, signature):
        semantics = self._get_semantics(chain_id, address)
        if semantics:
            transformations = semantics.contract.transformations.get(signature)
        else:
            transformations = None

        return transformations

    @lru_cache(maxsize=128)
    def get_anonymous_event_abi(self, chain_id, address):
        semantics = self._get_semantics(chain_id, address)
        event_semantics = None
        if semantics:
            anonymous_events = {
                signature
                for signature, event in semantics.contract.events.items()
                if event.anonymous
            }
            if len(anonymous_events) == 1:
                event_signature = anonymous_events.pop()
                event_semantics = semantics.contract.events[event_signature]

        return event_semantics

    @lru_cache(maxsize=128)
    def get_function_abi(self, chain_id, address, signature):
        semantics = self._get_semantics(chain_id, address)
        function_semantics = (
            semantics.contract.functions.get(signature) if semantics else None
        )

        return function_semantics

    @lru_cache(maxsize=128)
    def get_constructor_abi(self, chain_id, address):

        semantics = self._get_semantics(chain_id, address)
        constructor_semantics = (
            semantics.contract.functions.get("constructor") if semantics else None
        )
        if constructor_semantics:
            constructor_semantics.outputs.append(
                ParameterSemantics("__create_output__", "ignore", [], False, True)
            )

        return constructor_semantics

    def get_address_label(self, chain_id, address, token_proxies=None):
        semantics = self._get_semantics(chain_id, address)
        if semantics.erc20:
            contract_label = semantics.erc20.symbol
        elif token_proxies and address in token_proxies:
            contract_label = token_proxies[address][1] + "_proxy"
        else:
            contract_label = semantics.name if semantics and semantics.name else address

        return contract_label

    @lru_cache(maxsize=128)
    def check_is_contract(self, chain_id, address):
        semantics = self._get_semantics(chain_id, address)
        is_contract = semantics is not None and semantics.is_contract

        return is_contract

    @lru_cache(maxsize=128)
    def get_standard(self, chain_id, address):
        semantics = self._get_semantics(chain_id, address)
        standard = semantics.standard if semantics is not None else None

        return standard

    def get_token_data(self, chain_id, address, token_proxies=None):

        semantics = self._get_semantics(chain_id, address)
        if semantics and semantics.erc20:
            token_name = (
                semantics.erc20.name if semantics and semantics.erc20 else address
            )
            token_symbol = (
                semantics.erc20.symbol if semantics and semantics.erc20 else "Unknown"
            )
            token_decimals = (
                semantics.erc20.decimals if semantics and semantics.erc20 else 18
            )
        elif token_proxies and address in token_proxies:
            token_name, token_symbol, token_decimals = token_proxies[address]
        else:
            token_name = address
            token_symbol = "Unknown"
            token_decimals = 18

        return token_name, token_symbol, token_decimals

    def update_address(self, chain_id, address, contract):

        updated_address = {"network": chain_id, "address": address, **contract}
        self.database.insert_address(address_data=updated_address, update_if_exist=True)

        return updated_address

    def update_semantics(self, semantics):

        if not semantics:
            return

        address_semantics = semantics.json(False)
        contract_semantics = semantics.contract.json()

        self.database.insert_contract(contract_semantics, update_if_exist=True)
        self.database.insert_address(address_semantics, update_if_exist=True)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

from ..models.decoded_model import DecodedTransaction
from ..models.objects_model import Block, Call


class DecoderService:
    def __init__(self, abi_decoder, semantic_decoder, web3provider, default_chain):
        self.abi_decoder = abi_decoder
        self.semantic_decoder = semantic_decoder
        self.web3provider = web3provider
        self.default_chain = default_chain

    def get_delegations(self, call: Call) -> Dict[str, set]:
        delegations = {}

        if not call:
            return delegations

        calls_queue = [call]
        while calls_queue:

            call = calls_queue.pop()
            for _, sub_call in enumerate(call.subcalls):
                calls_queue.insert(0, sub_call)

            if call.call_type == "delegatecall":
                if call.from_address not in delegations:
                    delegations[call.from_address] = set()
                delegations[call.from_address].add(call.to_address)

        return delegations

    def get_token_proxies(self, delegations: Dict[str, set]) -> Dict[str, Dict]:
        token_proxies = {}

        for delegator in delegations:
            delegator_semantic = self.semantic_decoder.repository.get_token_data(
                self.default_chain, delegator
            )
            if (
                delegator_semantic[0] == delegator
                and delegator_semantic[1] == "Unknown"
            ):
                for delegate in delegations[delegator]:
                    delegate_semantic = self.semantic_decoder.repository.get_token_data(
                        self.default_chain, delegate
                    )
                    if (
                        delegate_semantic[0] != delegate
                        and delegate_semantic[1] != "Unknown"
                    ):
                        token_proxies[delegator] = delegate_semantic
                        break
            elif all(delegator_semantic):
                token_proxies[delegator] = delegator_semantic

        return token_proxies

    def decode_transaction(self, chain_id: str, tx_hash: str) -> DecodedTransaction:

        # verify the transaction hash
        tx_hash = tx_hash if tx_hash.startswith("0x") else "0x" + tx_hash

        chain_id = chain_id or self.default_chain

        # read a raw transaction from a node
        transaction = self.web3provider.get_full_transaction(
            tx_hash=tx_hash, chain_id=chain_id
        )
        # read a raw block from a node
        block = Block.from_raw(
            w3block=self.web3provider.get_block(transaction.metadata.block_number),
            chain_id=chain_id,
        )

        # prepare lists of delegations to properly decode delegate-calling contracts
        delegations = self.get_delegations(transaction.root_call)
        token_proxies = self.get_token_proxies(delegations)

        # decode transaction using ABI
        abi_decoded_tx = self.abi_decoder.decode_transaction(
            block=block,
            transaction=transaction,
            delegations=delegations,
            token_proxies=token_proxies,
            chain_id=chain_id,
        )

        # decode transaction using additional semantics
        semantically_decoded_tx = self.semantic_decoder.decode_transaction(
            block=block.metadata,
            transaction=abi_decoded_tx,
            token_proxies=token_proxies,
            chain_id=chain_id,
        )

        return semantically_decoded_tx



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import (
    EventSemantics,
    ParameterSemantics,
    TransformationSemantics,
    FunctionSemantics,
)

erc20_transfer_event = EventSemantics(
    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
    False,
    "Transfer",
    [
        ParameterSemantics("src", "address", [], True),
        ParameterSemantics("dst", "address", [], True),
        ParameterSemantics("value", "uint256", [], False),
    ],
)

erc20_transfer_event_transformation = {
    "__input2__": TransformationSemantics(
        transformation="__input2__ / 10**token_decimals(__contract__)"
    )
}

erc20_approval_event = EventSemantics(
    "0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925",
    False,
    "Approval",
    [
        ParameterSemantics("src", "address", [], True),
        ParameterSemantics("dst", "address", [], True),
        ParameterSemantics("value", "uint256", [], False),
    ],
)

erc20_approval_event_transformation = {
    "__input2__": TransformationSemantics(
        transformation="__input2__ / 10**token_decimals(__contract__)"
    )
}

erc20_transfer_function = FunctionSemantics(
    "0xa9059cbb",
    "transfer",
    [
        ParameterSemantics("recipient", "address", []),
        ParameterSemantics("amount", "uint256", []),
    ],
    [ParameterSemantics("", "bool", [])],
)

erc20_transfer_function_transformation = {
    "__input1__": TransformationSemantics(
        transformation="__input1__ / 10**token_decimals(__contract__)"
    )
}

erc20_transferFrom_function = FunctionSemantics(
    "0x23b872dd",
    "transferFrom",
    [
        ParameterSemantics("sender", "address", []),
        ParameterSemantics("recipient", "address", []),
        ParameterSemantics("amount", "uint256", []),
    ],
    [ParameterSemantics("", "bool", [])],
)

erc20_transferFrom_function_transformation = {
    "__input2__": TransformationSemantics(
        transformation="__input2__ / 10**token_decimals(__contract__)"
    )
}

erc20_approve_function = FunctionSemantics(
    "0x095ea7b3",
    "approve",
    [
        ParameterSemantics("spender", "address", []),
        ParameterSemantics("amount", "uint256", []),
    ],
    [ParameterSemantics("", "bool", [])],
)

erc20_approve_function_transformation = {
    "__input1__": TransformationSemantics(
        transformation="__input1__ / 10**token_decimals(__contract__)"
    )
}

erc20_balanceOf_function = FunctionSemantics(
    "0x70a08231",
    "balanceOf",
    [ParameterSemantics("holder", "address", [])],
    [ParameterSemantics("", "uint256", [])],
)

erc20_balanceOf_function_transformation = {
    "__output0__": TransformationSemantics(
        transformation="__output0__ / 10**token_decimals(__contract__)"
    )
}

erc20_totalSupply_function = FunctionSemantics(
    "0x18160ddd", "totalSupply", [], [ParameterSemantics("", "uint256", [])]
)

erc20_totalSupply_function_transformation = {
    "__output0__": TransformationSemantics(
        transformation="__output0__ / 10**token_decimals(__contract__)"
    )
}

ERC20_EVENTS = {
    erc20_transfer_event.signature: erc20_transfer_event,
    erc20_approval_event.signature: erc20_approval_event,
}

ERC20_FUNCTIONS = {
    erc20_transfer_function.signature: erc20_transfer_function,
    erc20_transferFrom_function.signature: erc20_transferFrom_function,
    erc20_approve_function.signature: erc20_approve_function,
    erc20_balanceOf_function.signature: erc20_balanceOf_function,
    erc20_totalSupply_function.signature: erc20_totalSupply_function,
}

ERC20_TRANSFORMATIONS = {
    erc20_transfer_event.signature: erc20_transfer_event_transformation,
    erc20_approval_event.signature: erc20_approval_event_transformation,
    erc20_transfer_function.signature: erc20_transfer_function_transformation,
    erc20_transferFrom_function.signature: erc20_transferFrom_function_transformation,
    erc20_approve_function.signature: erc20_approve_function_transformation,
    erc20_balanceOf_function.signature: erc20_balanceOf_function_transformation,
    erc20_totalSupply_function.signature: erc20_totalSupply_function_transformation,
}



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import partial


def token_decimals(transaction, repository, address):
    try:
        _, _, decimals = repository.get_token_data(transaction.chain_id, address)
    except:
        decimals = 18

    return decimals


def decode_nft(contract, token_id):
    if len(str(token_id)) > 8:
        token_symbol = f"NFT {str(token_id)[:6]}...{str(token_id)[-2:]}"
    else:
        token_symbol = f"NFT {token_id}"

    token_address = f"{contract}?a={token_id}#inventory"

    return dict(address=token_address, name=token_symbol)


def string_from_bytes(raw_value):
    try:
        raw_value = raw_value[2:] if raw_value[:2] == "0x" else raw_value
        decoded_string = bytes.fromhex(raw_value).decode("utf-8").replace("\x00", "")
    except:
        decoded_string = "???"

    return decoded_string


def add_utils_to_context(context):

    # register additional functions available for transformations
    context["token_decimals"] = partial(
        token_decimals, context["__transaction__"], context["__repository__"]
    )
    context["decode_nft"] = partial(decode_nft, context["__contract__"])
    context["string_from_bytes"] = string_from_bytes

    return



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.decoded_model import DecodedTransactionMetadata, AddressInfo
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata
from .abc import SemanticSubmoduleAbc
from .helpers.utils import get_eth_price


class SemanticMetadataDecoder(SemanticSubmoduleAbc):
    """Semantic Metadata Decoder."""

    def decode(
        self,
        block_metadata: BlockMetadata,
        tx_metadata: TransactionMetadata,
        chain_id: str,
    ) -> DecodedTransactionMetadata:
        """Semantically decode metadata."""

        decoded_metadata = DecodedTransactionMetadata(
            chain_id=chain_id,
            tx_hash=tx_metadata.tx_hash,
            block_number=block_metadata.block_number,
            timestamp=block_metadata.timestamp,
            gas_price=tx_metadata.gas_price / 10 ** 9,
            sender=AddressInfo(
                tx_metadata.from_address,
                self.repository.get_address_label(chain_id, tx_metadata.from_address),
                "sender",
            ),
            receiver=AddressInfo(
                tx_metadata.to_address,
                self.repository.get_address_label(chain_id, tx_metadata.to_address),
                "receiver",
            ),
            tx_index=tx_metadata.tx_index,
            tx_value=tx_metadata.tx_value,
            eth_price=get_eth_price(),
            gas_limit=tx_metadata.gas_limit,
            gas_used=tx_metadata.gas_used,
            success=tx_metadata.success,
        )

        return decoded_metadata



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

from ethtx.models.decoded_model import DecodedCall, DecodedTransactionMetadata
from ethtx.semantics.standards.erc20 import ERC20_FUNCTIONS, ERC20_TRANSFORMATIONS
from ethtx.semantics.standards.erc721 import ERC721_FUNCTIONS, ERC721_TRANSFORMATIONS
from ethtx.utils.measurable import RecursionLimit
from .abc import SemanticSubmoduleAbc
from .helpers.utils import (
    create_transformation_context,
    semantically_decode_parameter,
    get_badge,
)

RECURSION_LIMIT = 2000


class SemanticCallsDecoder(SemanticSubmoduleAbc):
    """Semantic Calls Decoder."""

    def decode(
        self,
        call: DecodedCall,
        tx_metadata: DecodedTransactionMetadata,
        token_proxies: Dict[str, Dict],
    ) -> DecodedCall:

        function_transformations = self.repository.get_transformations(
            call.chain_id, call.to_address.address, call.function_signature
        )

        if function_transformations:
            call.function_name = (
                function_transformations.get("name") or call.function_name
            )
        else:
            function_transformations = dict()

        # prepare context for transformations
        context = create_transformation_context(
            call.to_address.address,
            call.arguments,
            call.outputs,
            tx_metadata,
            self.repository,
        )
        standard = self.repository.get_standard(call.chain_id, call.to_address.address)

        # perform parameters transformations
        for i, parameter in enumerate(call.arguments):
            semantically_decode_parameter(
                self.repository,
                parameter,
                f"__input{i}__",
                function_transformations,
                token_proxies,
                context,
            )
        for i, parameter in enumerate(call.outputs):
            semantically_decode_parameter(
                self.repository,
                parameter,
                f"__output{i}__",
                function_transformations,
                token_proxies,
                context,
            )

        if standard == "ERC20":
            # decode ERC20 calls if transformations for them are not defined
            if call.function_signature in ERC20_FUNCTIONS and (
                not function_transformations
                or call.function_signature not in function_transformations
            ):
                function_transformations = ERC20_TRANSFORMATIONS.get(
                    call.function_signature
                )
                if function_transformations:
                    for i, parameter in enumerate(call.arguments):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__input{i}__",
                            function_transformations,
                            token_proxies,
                            context,
                        )
                    for i, parameter in enumerate(call.outputs):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__output{i}__",
                            function_transformations,
                            token_proxies,
                            context,
                        )
        elif standard == "ERC721":
            # decode ERC721 calls if transformations for them are not defined
            if call.function_signature in ERC721_FUNCTIONS and (
                not function_transformations
                or call.function_signature not in function_transformations
            ):
                function_transformations = ERC721_TRANSFORMATIONS.get(
                    call.function_signature
                )
                if function_transformations:
                    for i, parameter in enumerate(call.arguments):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__input{i}__",
                            function_transformations,
                            token_proxies,
                            context,
                        )
                    for i, parameter in enumerate(call.outputs):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__output{i}__",
                            function_transformations,
                            token_proxies,
                            context,
                        )

        # decode proper from and to addresses badges
        call.from_address.badge = get_badge(
            call.from_address.address, tx_metadata.sender, tx_metadata.receiver
        )
        call.to_address.badge = get_badge(
            call.to_address.address, tx_metadata.sender, tx_metadata.receiver
        )

        # remove ignored parameters
        call.arguments = [
            parameter for parameter in call.arguments if parameter.type != "ignore"
        ]
        call.outputs = [
            parameter for parameter in call.outputs if parameter.type != "ignore"
        ]

        with RecursionLimit(RECURSION_LIMIT):
            if call.subcalls:
                for sub_call in call.subcalls:
                    self.decode(sub_call, tx_metadata, token_proxies)

        return call



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Union, Dict

from web3 import Web3

from ethtx.decoders.semantic.helpers.utils import (
    create_transformation_context,
    semantically_decode_parameter,
    get_badge,
)
from ethtx.models.decoded_model import DecodedEvent, DecodedTransactionMetadata
from ethtx.semantics.protocols.anonymous import anonymous_events
from ethtx.semantics.standards.erc20 import ERC20_EVENTS, ERC20_TRANSFORMATIONS
from ethtx.semantics.standards.erc721 import ERC721_TRANSFORMATIONS, ERC721_EVENTS
from .abc import SemanticSubmoduleAbc

log = logging.getLogger(__name__)


class SemanticEventsDecoder(SemanticSubmoduleAbc):
    """Semantic Events Decoder."""

    def decode(
        self,
        events: Union[DecodedEvent, List[DecodedEvent]],
        tx_metadata: DecodedTransactionMetadata,
        token_proxies: Dict[str, Dict],
    ) -> Union[DecodedEvent, List[DecodedEvent]]:
        """Semantically decode events."""
        if isinstance(events, list):
            return (
                [
                    self.decode_event(event, tx_metadata, token_proxies)
                    for event in events
                ]
                if events
                else []
            )

        return self.decode_event(events, tx_metadata, token_proxies)

    def decode_event(
        self,
        event: DecodedEvent,
        tx_metadata: DecodedTransactionMetadata,
        token_proxies: Dict[str, Dict],
    ) -> DecodedEvent:
        """Semantically decode event"""

        if event.event_name != event.event_signature:
            # calculate signature to account for anonymous events
            parameters_types = [parameter.type for parameter in event.parameters]
            calculated_event_signature = Web3.keccak(
                text=f'{event.event_name}({",".join(parameters_types)})'
            ).hex()
        else:
            calculated_event_signature = event.event_signature

        if (
            event.event_name != event.event_signature
            and calculated_event_signature != event.event_signature
            and calculated_event_signature not in anonymous_events
        ):
            log.warning(
                "Event signature mismatch: %s / %s.",
                calculated_event_signature,
                event.event_signature,
            )

        # read transformations from the repository
        event_transformations = self.repository.get_transformations(
            event.chain_id, event.contract.address, calculated_event_signature
        )

        if event_transformations:
            event.event_name = event_transformations.get("name") or event.event_name
        else:
            if calculated_event_signature in anonymous_events and (
                not event_transformations
                or event.event_signature not in event_transformations
            ):
                event_transformations = anonymous_events[calculated_event_signature]
            else:
                event_transformations = dict()

        # prepare context for transformations
        context = create_transformation_context(
            event.contract.address, event.parameters, [], tx_metadata, self.repository
        )
        standard = self.repository.get_standard(event.chain_id, event.contract.address)
        if not standard and event.contract.address in token_proxies:
            standard = "ERC20"

        # perform parameters transformation
        for i, parameter in enumerate(event.parameters):
            semantically_decode_parameter(
                self.repository,
                parameter,
                f"__input{i}__",
                event_transformations,
                token_proxies,
                context,
            )

        if standard == "ERC20":
            # decode ERC20 events if transformations for them are not defined
            if event.event_signature in ERC20_EVENTS and (
                not event_transformations
                or event.event_signature not in event_transformations
            ):
                event_transformations = ERC20_TRANSFORMATIONS.get(event.event_signature)
                if event_transformations:
                    for i, parameter in enumerate(event.parameters):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__input{i}__",
                            event_transformations,
                            token_proxies,
                            context,
                        )
        elif standard == "ERC721":
            # decode ERC721 events if transformations for them are not defined
            if event.event_signature in ERC721_EVENTS and (
                not event_transformations
                or event.event_signature not in event_transformations
            ):
                event_transformations = ERC721_TRANSFORMATIONS.get(
                    event.event_signature
                )
                if event_transformations:
                    for i, parameter in enumerate(event.parameters):
                        semantically_decode_parameter(
                            self.repository,
                            parameter,
                            f"__input{i}__",
                            event_transformations,
                            token_proxies,
                            context,
                        )

        # decode proper emitter badge
        event.contract.badge = get_badge(
            event.contract.address, tx_metadata.sender, tx_metadata.receiver
        )

        # remove ignored parameters
        event.parameters = [
            parameter for parameter in event.parameters if parameter.type != "ignore"
        ]

        return event



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import traceback
from typing import Optional, Dict, List

from ethtx.models.decoded_model import (
    DecodedTransaction,
    DecodedCall,
    DecodedEvent,
    DecodedTransfer,
)
from ethtx.models.objects_model import (
    Block,
    BlockMetadata,
    Transaction,
    TransactionMetadata,
    Call,
    Event,
)
from ethtx.utils.measurable import ExecutionTimer
from .abc import IABIDecoder
from .balances import ABIBalancesDecoder
from .calls import ABICallsDecoder
from .events import ABIEventsDecoder
from .transfers import ABITransfersDecoder

log = logging.getLogger(__name__)


class ABIDecoder(IABIDecoder):
    def decode_transaction(
        self,
        block: Block,
        transaction: Transaction,
        chain_id: str,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> Optional[DecodedTransaction]:

        log.info("ABI decoding for %s / %s.", transaction.metadata.tx_hash, chain_id)

        try:
            with ExecutionTimer(f"ABI decoding for " + transaction.metadata.tx_hash):
                log.info(
                    "ABI decoding for %s / %s.", transaction.metadata.tx_hash, chain_id
                )
                full_decoded_transaction = self._decode_transaction(
                    block.metadata, transaction, chain_id, delegations, token_proxies
                )
                return full_decoded_transaction
        except Exception as e:
            log.warning(
                "ABI decoding of %s / %s failed.",
                transaction.metadata.tx_hash,
                chain_id,
            )
            traceback.print_exc(e)
            return None

    def decode_calls(
        self,
        root_call: Call,
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> Optional[DecodedCall]:
        return ABICallsDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(
            call=root_call,
            transaction=transaction,
            delegations=delegations,
            token_proxies=token_proxies,
        )

    def decode_call(
        self,
        root_call: Call,
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> Optional[DecodedCall]:
        return ABICallsDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(
            call=root_call,
            transaction=transaction,
            delegations=delegations,
            token_proxies=token_proxies,
        )

    def decode_events(
        self,
        events: [Event],
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> List[DecodedEvent]:
        return ABIEventsDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(
            events=events,
            transaction=transaction,
            delegations=delegations or {},
            token_proxies=token_proxies or {},
        )

    def decode_event(
        self,
        events: Event,
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> DecodedEvent:
        return ABIEventsDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(
            events=events,
            transaction=transaction,
            delegations=delegations or {},
            token_proxies=token_proxies or {},
        )

    def decode_transfers(
        self,
        call: DecodedCall,
        events: List[DecodedEvent],
        token_proxies: Optional[Dict[str, dict]] = None,
    ):
        return ABITransfersDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(call=call, events=events, token_proxies=token_proxies or {})

    def decode_balances(self, transfers: List[DecodedTransfer]):
        return ABIBalancesDecoder(
            repository=self._repository, chain_id=self._default_chain
        ).decode(transfers=transfers)

    def _decode_transaction(
        self,
        block: BlockMetadata,
        transaction: Transaction,
        chain_id: str,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
    ) -> DecodedTransaction:

        full_decoded_transaction = DecodedTransaction(
            block_metadata=block,
            tx_metadata=transaction.metadata,
            events=[],
            calls=None,
            transfers=[],
            balances=[],
        )

        self._repository.record()

        try:
            full_decoded_transaction.calls = self.decode_calls(
                transaction.root_call, transaction.metadata, delegations, token_proxies
            )
        except Exception as e:
            log.warning(
                "ABI decoding of calls tree for %s / %s failed.",
                transaction.metadata.tx_hash,
                chain_id,
            )
            log.warning(e)
            return full_decoded_transaction

        try:
            full_decoded_transaction.events = self.decode_events(
                transaction.events, transaction.metadata, delegations, token_proxies
            )
        except Exception as e:
            log.warning(
                "ABI decoding of events for %s / %s failed.",
                transaction.metadata.tx_hash,
                chain_id,
            )
            log.warning(e)
            return full_decoded_transaction

        try:
            full_decoded_transaction.transfers = self.decode_transfers(
                full_decoded_transaction.calls,
                full_decoded_transaction.events,
                token_proxies,
            )
        except Exception as e:
            log.warning(
                "ABI decoding of transfers for %s / %s failed.",
                transaction.metadata.tx_hash,
                chain_id,
            )
            log.warning(e)
            return full_decoded_transaction

        try:
            full_decoded_transaction.balances = self.decode_balances(
                full_decoded_transaction.transfers
            )
        except Exception as e:
            log.warning(
                "ABI decoding of balances for %s / %s failed.",
                transaction.metadata.tx_hash,
                chain_id,
            )
            log.warning(e)
            return full_decoded_transaction

        used_semantics = self._repository.end_record()
        log.info(
            f"Semantics used in decoding {transaction.metadata.tx_hash}: "
            + ", ".join(used_semantics)
        )

        full_decoded_transaction.status = True

        return full_decoded_transaction



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from datetime import datetime

from ethtx.models.decoded_model import Argument
from ethtx.models.semantics_model import ParameterSemantics

log = logging.getLogger(__name__)


def decode_event_parameters(data, topics, abi, anonymous):
    # for anonymous events shift the list of topics
    if anonymous:
        topics.insert(0, None)

    parameters_abi = abi.parameters if abi else []

    topic_parameters = {}
    data_parameters = {}
    idx_parameters = []
    data = data[2:]

    if parameters_abi:

        idx_parameters = [p.indexed or False for p in parameters_abi]

        # parse indexed parameters
        for i, parameter in enumerate([p for p in parameters_abi if p.indexed]):
            parameter_name = parameter.parameter_name
            parameter_type = parameter.parameter_type
            # assumption that topic parameters can be only static
            if len(topics) > i + 1:
                raw_parameter = (
                    topics[i + 1]
                    if isinstance(topics[i + 1], str)
                    else topics[i + 1].hex()
                )
                if not parameter_type:
                    parameter_value = raw_parameter
                    parameter_type = "unknown"
                else:
                    parameter_value = decode_static_argument(
                        raw_parameter, parameter_type
                    )
                topic_parameters[i] = Argument(
                    parameter_name, parameter_type, parameter_value
                )
            else:
                log.warning("Topics length mismatch.")
                return []

        # parse other (not indexed) parameters
        if len(data):
            parameters, _ = decode_struct(
                data, [p for p in parameters_abi if not p.indexed]
            )
            for (i, parameter) in enumerate(parameters):
                data_parameters[i] = Argument(**parameter)

    else:

        for i, parameter in enumerate(topics[1:]):
            if not topics[i + 1]:
                break
            parameter_name = ""
            parameter_type = "unknown"
            parameter_value = (
                topics[i + 1] if isinstance(topics[i + 1], str) else topics[i + 1].hex()
            )

            topic_parameters[i] = Argument(
                parameter_name, parameter_type, parameter_value
            )

        no_parameters = len(data) // 64
        for i in range(no_parameters):
            parameter_name = ""
            parameter_type = "unknown"
            parameter_value = data[64 * i : 64 * (i + 1)]

            data_parameters[i] = Argument(
                parameter_name, parameter_type, parameter_value
            )

    # store parameters in original ABI order
    event_parameters = []
    if parameters_abi:
        ni = nd = 0
        for i in range(len(parameters_abi)):
            if idx_parameters[i]:
                event_parameters.append(topic_parameters[ni])
                ni += 1
            else:
                event_parameters.append(data_parameters[nd])
                nd += 1
    else:
        for _, parameter in topic_parameters.items():
            event_parameters.append(parameter)
        for _, parameter in data_parameters.items():
            event_parameters.append(parameter)

    return event_parameters


def decode_function_parameters(input_data, output, abi, status=True):
    if abi:
        input_parameters, _ = decode_struct(input_data[10:], abi.inputs)
        for i, parameter in enumerate(input_parameters):
            input_parameters[i] = Argument(**parameter)
    elif input_data[10:]:
        input_parameters = [
            Argument(name="call_data", type="bytes", value="0x" + input_data[10:])
        ]
    else:
        input_parameters = []

    if not status and output[:10] == "0x08c379a0":
        error_abi = ParameterSemantics("Error", "string", [], False, True)
        error_parameters, _ = decode_struct(output[10:], [error_abi])
        output_parameters = [Argument(**error_parameters[0])]
    else:
        if abi:
            if abi.outputs and output == "0x":
                log.warning("Warning: missing output data...")
                output_parameters = []
            else:
                output_parameters, _ = decode_struct(output[2:], abi.outputs)
                for i, parameter in enumerate(output_parameters):
                    output_parameters[i] = Argument(**parameter)
        elif output != "0x":
            output_parameters = [
                Argument(name="output_data", type="bytes", value=output)
            ]
        else:
            output_parameters = []

    return input_parameters, output_parameters


# helper function to decode an argument value based on expected type
def decode_static_argument(raw_value, argument_type):
    decoded_value = raw_value

    if decoded_value:

        if argument_type == "address":
            if len(raw_value) >= 40:
                decoded_value = "0x" + raw_value[-40:]
            else:
                decoded_value = raw_value

        elif argument_type[:4] == "uint":
            if isinstance(raw_value, str):
                decoded_value = int(raw_value, 16)
            else:
                decoded_value = raw_value

        elif argument_type[:3] == "int":
            if isinstance(raw_value, str):
                decoded_value = int(raw_value, 16)
                if decoded_value & (1 << (256 - 1)):
                    decoded_value -= 1 << 256
            else:
                decoded_value = raw_value

        elif argument_type == "bool":
            if int(raw_value, 16) == 0:
                decoded_value = "False"
            else:
                decoded_value = "True"

        elif argument_type == "bytes":
            decoded_value = "0x" + bytes.fromhex(raw_value[2:]).hex()

        elif argument_type[:5] == "bytes":
            decoded_value = "0x" + raw_value

        elif argument_type == "byte":
            decoded_value = "0x" + bytes.fromhex(raw_value[2:])[0].hex()

        elif argument_type in ("string", "string32"):
            try:
                if raw_value[:2] == "0x":
                    raw_value = raw_value[2:]
                decoded_value = (
                    bytes.fromhex(raw_value).decode("utf-8").replace("\x00", "")
                )
            except Exception:
                pass

        elif argument_type == "timestamp":
            if isinstance(raw_value, str):
                decoded_value = str(datetime.utcfromtimestamp(int(raw_value, 16)))
            else:
                decoded_value = str(datetime.utcfromtimestamp(raw_value))

        elif argument_type == "hashmap":
            decoded_value = "[...]"
        elif argument_type == "tuple":
            decoded_value = "(...)"
        elif argument_type == "tuple[]":
            decoded_value = "(...)[]"

    return decoded_value


# helper function to decode ABI 2.0
def decode_tuple(data, argument_abi, is_list):
    slots = 0

    if is_list:
        count = int(data[:64], 16)
        data = data[64:]
        decoded_argument = []

        for c in range(count):
            do_offset = any(a.dynamic for a in argument_abi)
            if do_offset:
                raw_value = data[c * 64 : (c + 1) * 64]
                offset = int(raw_value, 16) * 2
                sub_bytes = data[offset:]
            else:
                sub_bytes = data

            decoded, num = decode_struct(sub_bytes, argument_abi)
            for i, parameter in enumerate(decoded):
                decoded[i] = Argument(**parameter)
            decoded_argument.append(decoded)
            slots += num

    else:
        decoded_argument, num = decode_struct(data, argument_abi)
        for i, parameter in enumerate(decoded_argument):
            decoded_argument[i] = Argument(**parameter)
        slots += num

    return decoded_argument, slots


# helper function to decode dynamic arrays
def decode_dynamic_array(data, array_type):
    count = int(data[:64], 16)
    sub_bytes = data[64:]
    decoded_argument = []

    for _ in range(count):
        if array_type in ("bytes", "string"):
            offset = int(sub_bytes[:64], 16) * 2
            decoded = decode_dynamic_argument(sub_bytes[offset:], array_type)
        else:
            decoded = decode_static_argument(sub_bytes[:64], array_type)

        decoded_argument.append(decoded)
        sub_bytes = sub_bytes[64:]

    return decoded_argument


# helper function to decode a dynamic argument
def decode_dynamic_argument(argument_bytes, argument_type):
    if len(argument_bytes):
        length = int(argument_bytes[:64], 16) * 2
        value = argument_bytes[64 : 64 + length]

        if argument_type == "string":
            decoded_value = (
                bytes.fromhex(value).decode("utf-8").replace("\x00", "").encode()
            )
        else:
            decoded_value = "0x" + value
    else:
        decoded_value = bytes(0).decode()

    return decoded_value


# helper function to decode ABI 2.0 structs
def decode_struct(data, arguments_abi):
    if arguments_abi:
        no_arguments = len(arguments_abi)
    else:
        no_arguments = len(data) // 64 + 1

    arguments_list = []
    slot = 0
    for i in range(no_arguments):
        raw_value = data[slot * 64 : (slot + 1) * 64]

        if arguments_abi:

            argument_name = arguments_abi[i].parameter_name
            argument_type = arguments_abi[i].parameter_type

            if argument_type[:5] == "tuple":
                do_offset = arguments_abi[i].dynamic or any(
                    a.dynamic for a in arguments_abi[i].components
                )
                if do_offset:
                    offset = int(raw_value, 16) * 2
                    sub_arguments = data[offset:]
                else:
                    sub_arguments = data[i * 64 :]

                argument_value, slots = decode_tuple(
                    sub_arguments,
                    arguments_abi[i].components,
                    argument_type[5:] == "[]",
                )

                if do_offset:
                    slot += 1
                else:
                    slot += slots

            elif argument_type in ("bytes", "string"):
                offset = int(raw_value, 16) * 2
                argument_value = decode_dynamic_argument(data[offset:], argument_type)
                slot += 1

            elif argument_type[-1:] == "]":
                array_type = argument_type[:-1].split("[")[0]

                if argument_type[-2:] == "[]":
                    offset = int(raw_value, 16) * 2
                    argument_value = decode_dynamic_array(data[offset:], array_type)
                    slot += 1
                else:
                    array_size = int(argument_type[:-1].split("[")[1])
                    argument_value = []
                    for _ in range(array_size):
                        argument_value.append(
                            decode_static_argument(raw_value, array_type)
                        )
                        slot += 1
                        raw_value = data[slot * 64 : (slot + 1) * 64]

            else:
                argument_value = decode_static_argument(raw_value, argument_type)
                slot += 1
        else:
            argument_name = "arg_%d" % (i + 1)
            argument_type = "unknown"
            argument_value = "0x" + raw_value

        if argument_type != "unknown" or argument_value != "0x":
            arguments_list.append(
                dict(name=argument_name, type=argument_type, value=argument_value)
            )

    return arguments_list, slot


def decode_graffiti_parameters(input_data):
    input_parameters = []

    if input_data and len(input_data) > 2:
        try:
            message = bytearray.fromhex(input_data[2:]).decode()
            input_parameters = [Argument(name="message", type="string", value=message)]
        except Exception as e:
            log.warning(e)

    return input_parameters



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, List, Dict

from ethtx.models.decoded_model import (
    DecodedTransaction,
    DecodedCall,
    DecodedEvent,
    DecodedTransfer,
    DecodedBalance,
)
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata
from ethtx.providers.semantic_providers.semantics_repository import SemanticsRepository


class SemanticSubmoduleAbc(ABC):
    def __init__(self, repository: SemanticsRepository):
        self.repository = repository

    @abstractmethod
    def decode(self, *args, **kwargs) -> Any:
        ...


class ISemanticDecoder(ABC):
    def __init__(self, repository: SemanticsRepository, chain_id: str):
        self.repository = repository
        self._default_chain = chain_id

    @abstractmethod
    def decode_transaction(
        self,
        block: BlockMetadata,
        transaction: DecodedTransaction,
        token_proxies: Dict[str, Dict],
        chain_id: str,
    ) -> DecodedTransaction:
        ...

    @abstractmethod
    def decode_metadata(
        self, block_metadata: BlockMetadata, tx_metadata: TransactionMetadata
    ):
        ...

    @abstractmethod
    def decode_calls(
        self,
        call: DecodedCall,
        tx_metadata: TransactionMetadata,
        token_proxies: Dict[str, Dict],
    ) -> SemanticSubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_events(
        self,
        events: List[DecodedEvent],
        tx_metadata: TransactionMetadata,
        token_proxies: Dict[str, Dict],
    ) -> SemanticSubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_transfers(
        self, transfers: List[DecodedTransfer]
    ) -> SemanticSubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_balances(
        self, balances: List[DecodedBalance]
    ) -> SemanticSubmoduleAbc.decode:
        ...



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Tuple

from ethtx.models.semantics_model import (
    ParameterSemantics,
    EventSemantics,
    FunctionSemantics,
    TransformationSemantics,
)


def _decode_parameters_list(raw_parameters_list: list) -> List[ParameterSemantics]:
    parameters_list = []

    if not raw_parameters_list:
        return parameters_list

    for raw_parameter_semantics in raw_parameters_list:

        if "indexed" in raw_parameter_semantics:
            indexed = raw_parameter_semantics["indexed"]
        else:
            indexed = False

        if "dynamic" in raw_parameter_semantics:
            dynamic = raw_parameter_semantics["dynamic"]
        else:
            dynamic = False

        if raw_parameter_semantics["type"] == "tuple":
            components = _decode_parameters_list(raw_parameter_semantics["components"])
        else:
            components = []

        parameters_list.append(
            ParameterSemantics(
                raw_parameter_semantics["name"],
                raw_parameter_semantics["type"],
                components,
                indexed,
                dynamic,
            )
        )
    return parameters_list


def decode_events_and_functions(
    abi: dict,
) -> Tuple[Dict[str, EventSemantics], Dict[str, FunctionSemantics]]:
    events = dict()
    for signature, raw_event_semantics in abi.get("events", {}).items():
        parameters = _decode_parameters_list(raw_event_semantics.get("parameters"))
        events[signature] = EventSemantics(
            signature,
            raw_event_semantics["anonymous"],
            raw_event_semantics["name"],
            parameters,
        )

    functions = dict()
    for signature, raw_function_semantics in abi.get("functions", {}).items():
        if raw_function_semantics:
            inputs = _decode_parameters_list(raw_function_semantics.get("inputs"))
            outputs = _decode_parameters_list(raw_function_semantics.get("outputs"))
            name = raw_function_semantics["name"]
        else:
            inputs = outputs = []
            name = signature

        functions[signature] = FunctionSemantics(signature, name, inputs, outputs)

    return events, functions


def decode_transformations(
    raw_transformations: dict,
) -> Dict[str, Dict[str, TransformationSemantics]]:
    transformations = dict()
    if raw_transformations:
        for signature, transformation in raw_transformations.items():
            transformations[signature] = dict()
            for parameter_name, parameter_transformation in transformation.get(
                "arguments", dict()
            ).items():
                transformations[signature][parameter_name] = TransformationSemantics(
                    parameter_transformation.get("name"),
                    parameter_transformation.get("type"),
                    parameter_transformation.get("value"),
                )
    return transformations



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from hexbytes import HexBytes

from ethtx.models.objects_model import BlockMetadata, TransactionMetadata, Event, Call


@dataclass
class W3Block:
    chain_id: str
    difficulty: int
    extraData: HexBytes
    gasLimit: int
    gasUsed: int
    hash: HexBytes
    logsBloom: HexBytes
    miner: str
    nonce: HexBytes
    number: int
    parentHash: HexBytes
    receiptsRoot: HexBytes
    sha3Uncles: HexBytes
    size: int
    stateRoot: HexBytes
    timestamp: int
    totalDifficulty: int
    transactions: List
    transactionsRoot: HexBytes
    uncles: List

    def to_object(self) -> BlockMetadata:
        block_hash = self.hash.hex()
        timestamp = datetime.utcfromtimestamp(self.timestamp)
        parent_hash = self.parentHash.hex()
        miner = self.miner.lower()
        gas_limit = self.gasLimit
        gas_used = self.gasUsed
        tx_count = len(self.transactions)

        return BlockMetadata(
            block_number=self.number,
            block_hash=block_hash,
            timestamp=timestamp,
            parent_hash=parent_hash,
            miner=miner,
            gas_limit=gas_limit,
            gas_used=gas_used,
            tx_count=tx_count,
        )


@dataclass
class W3Transaction:
    chain_id: str
    blockHash: str
    blockNumber: int
    from_address: str
    gas: int
    gasPrice: int
    hash: HexBytes
    input: str
    nonce: int
    r: HexBytes
    s: HexBytes
    to: str
    transactionIndex: int
    v: int
    value: int

    def to_object(self, w3receipt: W3Receipt) -> TransactionMetadata:
        tx_hash = self.hash.hex()
        block_number = self.blockNumber
        tx_index = self.transactionIndex
        from_address = self.from_address.lower()
        to_address = self.to.lower() if self.to else None
        tx_value = self.value
        gas_limit = self.gas
        gas_price = self.gasPrice
        gas_used = w3receipt.gasUsed
        success = w3receipt.status == 1

        return TransactionMetadata(
            tx_hash=tx_hash,
            block_number=block_number,
            tx_index=tx_index,
            from_address=from_address,
            to_address=to_address,
            tx_value=tx_value,
            gas_limit=gas_limit,
            gas_price=gas_price,
            gas_used=gas_used,
            success=success,
        )


@dataclass
class W3Receipt:
    tx_hash: str
    chain_id: str
    blockHash: HexBytes
    blockNumber: int
    contractAddress: str
    cumulativeGasUsed: int
    from_address: str
    gasUsed: int
    logsBloom: HexBytes
    root: str
    status: int
    to_address: str
    transactionHash: HexBytes
    transactionIndex: int
    logs: list = field(default_factory=list)


@dataclass
class W3Log:
    tx_hash: str
    chain_id: str
    address: str
    blockHash: HexBytes
    blockNumber: int
    data: str
    logIndex: int
    removed: bool
    topics: List[HexBytes]
    transactionHash: HexBytes
    transactionIndex: int

    def to_object(self) -> Event:
        contract = self.address.lower()
        log_index = self.logIndex
        log_data = self.data
        topics = []

        for i in range(len(self.topics)):
            topics.append(self.topics[i].hex())

        return Event(
            contract=contract, topics=topics, log_data=log_data, log_index=log_index
        )


@dataclass
class W3CallTree:
    tx_hash: str
    chain_id: str
    type: str
    from_address: str
    to_address: str
    input: str
    output: str
    value: str = None
    time: str = None
    gas: str = None
    gasUsed: str = None
    error: str = None
    calls: list = field(default_factory=list)

    def to_object(self) -> Call:
        from_address = self.from_address
        to_address = self.to_address
        call_value = int(self.value, 16) if self.value else 0
        call_type = self.type.lower()
        call_data = self.input
        return_value = self.output
        gas_used = int(self.gasUsed, 16) if self.gasUsed else None
        call_gas = int(self.gas, 16) if self.gas else None
        status = self.error is None
        error = self.error

        call = Call(
            call_type=call_type,
            from_address=from_address,
            to_address=to_address,
            call_value=call_value,
            call_data=call_data,
            return_value=return_value,
            gas_used=gas_used,
            call_gas=call_gas,
            status=status,
            error=error,
        )

        for child_call in self.calls:
            call.subcalls.append(child_call.to_object())

        return call



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

from mongoengine import connect
from pymongo import MongoClient

from .decoders.abi.decoder import ABIDecoder
from .decoders.decoder_service import DecoderService
from .decoders.semantic.decoder import SemanticDecoder
from .models.objects_model import Call
from .providers.etherscan_provider import EtherscanProvider
from .providers.semantic_providers.semantics_database import (
    MongoSemanticsDatabase,
    ISemanticsDatabase,
)
from .providers.semantic_providers.semantics_repository import SemanticsRepository
from .providers.web3_provider import Web3Provider
from .utils.validators import assert_tx_hash


class EthTxConfig:
    mongo_connection_string: str
    etherscan_api_key: str
    web3nodes: Dict[str, str]
    mongo_database: str
    etherscan_urls: Dict[str, str]
    default_chain: str

    def __init__(
        self,
        mongo_connection_string: str,
        mongo_database: str,
        web3nodes: Dict[str, str],
        etherscan_api_key: str,
        etherscan_urls: Dict[str, str],
        default_chain: str = "mainnet",
    ):
        self.mongo_connection_string = mongo_connection_string
        self.etherscan_api_key = etherscan_api_key
        self.web3nodes = web3nodes
        self.mongo_database = mongo_database
        self.default_chain = default_chain
        self.etherscan_urls = etherscan_urls


class EthTxDecoders:
    semantic_decoder: SemanticDecoder
    abi_decoder: ABIDecoder

    def __init__(
        self,
        semantic_decoder: SemanticDecoder,
        abi_decoder: ABIDecoder,
        decoder_service: DecoderService,
    ):
        self.semantic_decoder = semantic_decoder
        self.abi_decoder = abi_decoder
        self._decoder_service = decoder_service

    def decode_transaction(self, tx_hash: str, chain_id: str = None):
        assert_tx_hash(tx_hash)
        return self._decoder_service.decode_transaction(chain_id, tx_hash)

    def get_proxies(self, call_tree: Call):
        delegations = self._decoder_service.get_delegations(call_tree)
        return self._decoder_service.get_token_proxies(delegations)


class EthTxProviders:
    web3provider: Web3Provider

    def __init__(self, web3provider: Web3Provider):
        self.web3provider = web3provider


class EthTx:
    @staticmethod
    def initialize(config: EthTxConfig):
        default_chain = config.default_chain
        mongo_client: MongoClient = connect(
            config.mongo_database, host=config.mongo_connection_string
        )
        repository = MongoSemanticsDatabase(mongo_client.db)
        web3provider = Web3Provider(
            config.web3nodes, default_chain=config.default_chain
        )
        etherscan = EtherscanProvider(
            config.etherscan_api_key,
            config.etherscan_urls,
            default_chain_id=config.default_chain,
        )

        return EthTx(default_chain, web3provider, repository, etherscan)

    semantics: SemanticsRepository

    def __init__(
        self,
        default_chain: str,
        web3provider: Web3Provider,
        repository: ISemanticsDatabase,
        etherscan: EtherscanProvider,
    ):
        self._default_chain = default_chain
        self._semantics = SemanticsRepository(repository, etherscan, web3provider)
        abi_decoder = ABIDecoder(self.semantics, self._default_chain)
        semantic_decoder = SemanticDecoder(self.semantics, self._default_chain)
        decoder_service = DecoderService(
            abi_decoder, semantic_decoder, web3provider, self._default_chain
        )
        self._decoders = EthTxDecoders(semantic_decoder, abi_decoder, decoder_service)
        self._providers = EthTxProviders(web3provider)

    @property
    def decoders(self) -> EthTxDecoders:
        """EthTx Decoders."""
        return self._decoders

    @property
    def semantics(self) -> SemanticsRepository:
        """EthTx Semantics Repository."""
        return self._semantics

    @property
    def providers(self) -> EthTxProviders:
        """EthTx Providers."""
        return self._providers

    @property
    def default_chain(self) -> str:
        """Default chain."""
        return self._default_chain

    @default_chain.setter
    def default_chain(self, chain: str) -> None:
        """Default chain setter."""
        self._default_chain = chain



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# root dir
root = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = "EthTx"
DESCRIPTION = "EthTx transaction decoder."
URL = "https://github.com/EthTx/ethtx"
EMAIL = "tomek@tfi.team, karol@tfi.team, piotr.rudnik@tfi.team"
AUTHOR = "Tomasz Mierzwa, Karol Chojnowski, Piotr Rudnik"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = ""

REQUIRED = []
REQUIRED_TEST = []

about = {}
if not VERSION:
    with open(os.path.join("ethtx", "_version.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                _, _, version = line.replace('"', "").split()
                about["__version__"] = version
                break
else:
    about["__version__"] = VERSION

try:
    with io.open(os.path.join(root, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


def load_requirements(fname):
    """Load requirements from file."""
    with open(fname) as file:
        return file.read().splitlines()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(root, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# *************** INSTALL *****************
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    license="Apache-2.0 License",
    packages=find_packages(exclude=["tests"]),
    tests_require=load_requirements("requirements_test.txt"),
    install_requires=load_requirements("requirements.txt"),
    include_package_data=True,
    test_suite="tests",
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


__all__ = ["Web3ConnectionException", "ProcessingException", "InvalidTransactionHash"]


class Web3ConnectionException(Exception):
    """Web3 Connection Exception."""

    def __init__(self):
        super().__init__("Couldn't connect to web3provider")


class ProcessingException(Exception):
    """Processing Exception."""

    def __init__(self, msg):
        super().__init__("Exception processing: " + msg)


class InvalidTransactionHash(Exception):
    """Invalid Transaction Hash."""

    def __init__(self, tx_hash):
        super().__init__("Invalid transaction hash provided: " + tx_hash)



from typing import Dict

from hexbytes import HexBytes

from ethtx.models.w3_model import W3Transaction, W3Receipt, W3Block, W3Log
from ethtx.utils.attr_dict import AttrDict


class MockWeb3Provider:
    blocks = {
        1: {
            "difficulty": 123,  # int
            "extraData": "test",  # HexBytes
            "gasLimit": 123,  # int
            "gasUsed": 123,  # int
            "hash": HexBytes(
                b"\x88\xe9mE7\xbe\xa4\xd9\xc0]\x12T\x99\x07\xb3%a\xd3\xbf1\xf4Z\xaesL\xdc\x11\x9f\x13@l\xb6"
            ),  # str
            "logsBloom": "test",  # HexBytes
            "miner": "test",  # str
            "nonce": "test",  # HexBytes
            "number": 123,  # int
            "parentHash": HexBytes(
                b"\x88\xe9mE7\xbe\xa4\xd9\xc0]\x12T\x99\x07\xb3%a\xd3\xbf1\xf4Z\xaesL\xdc\x11\x9f\x13@l\xb6"
            ),  # str
            "receiptsRoot": "test",  # HexBytes
            "sha3Uncles": "test",  # HexBytes
            "size": 123,  # int
            "stateRoot": "test",  # HexBytes
            "timestamp": 123,  # int,
            "totalDifficulty": 123,  # int
            "transactions": [],  # List
            "transactionsRoot": "test",  # HexBytes
            "uncles": [],  # List
        }
    }

    txs = {
        "0xd7701a0fc05593aee3a16f20cab605db7183f752ae942cc75fd0975feaf1072e": {
            "blockHash": HexBytes(
                b"\x88\xe9mE7\xbe\xa4\xd9\xc0]\x12T\x99\x07\xb3%a\xd3\xbf1\xf4Z\xaesL\xdc\x11\x9f\x13@l\xb6"
            ),  # str
            "blockNumber": 1,  # int
            "from_address": "fromasd",  # str
            "gas": 420,  # int
            "gasPrice": 1,  # int
            "hash": HexBytes(
                b"\x88\xe9mE7\xbe\xa4\xd9\xc0]\x12T\x99\x07\xb3%a\xd3\xbf1\xf4Z\xaesL\xdc\x11\x9f\x13@l\xb6"
            ),  # HexBytes,
            "input": "jeszcze jak",  # str
            "nonce": 123,  # int
            "r": "ds",  # HexBytes
            "s": "sdf",  # HexBytes
            "to": "sdf",  # str
            "transactionIndex": 1,  # int
            "v": 1,  # int
            "value": 1,  # int
        }
    }

    def add_mocked_block_details(self, block_number, block_details: Dict):
        self.blocks[block_number] = block_details

    def get_transaction(self, tx_hash, chain_id="mainnet"):
        return W3Transaction(chain_id=chain_id, **self.txs[tx_hash])

    def get_receipt(self, tx_hash, chain_id):
        log_values = AttrDict(
            {
                "tx_hash": tx_hash,
                "chain_id": chain_id,
                "address": "test",  # str
                "blockHash": "test",  # HexBytes
                "blockNumber": 123,  # int
                "data": "test",  # str
                "logIndex": 132,  # int
                "removed": False,  # bool,
                "topics": [HexBytes("d")],  # List[HexBytes]
                "transactionHash": "test",  # HexBytes
                "transactionIndex": 123,  # int
            }
        )

        log = W3Log(**log_values)
        values = {
            "blockHash": "test",  # HexBytes
            "blockNumber": 123,  # int
            "contractAddress": 123,  # str
            "cumulativeGasUsed": 132,  # int,
            "from_address": "from",  # str
            "gasUsed": 123,  # int
            "logs": [log],  # List
            "logsBloom": "test",  # HexBytes
            "root": "test",  # str
            "status": 123,  # int,
            "to_address": "test",  # str
            "transactionHash": "test",  # HexBytes
            "transactionIndex": 123,  # int
        }
        return W3Receipt(tx_hash, chain_id, **values)

    def get_block(self, block_number: int, chain_id: str = None) -> W3Block:
        return W3Block(chain_id=chain_id, **self.blocks[block_number])



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from ethtx.exceptions import InvalidTransactionHash


def assert_tx_hash(tx_hash) -> None:
    tx_hash_regex = "^(0x)?([A-Fa-f0-9]{64})$"
    if not re.match(tx_hash_regex, tx_hash):
        raise InvalidTransactionHash(tx_hash)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime

import requests
from ethereum.transactions import Transaction


def decode_from_4byte(method_sig, decoded_methods):
    if method_sig not in decoded_methods:
        url = (
            "https://www.4byte.directory/api/v1/signatures/?hex_signature=" + method_sig
        )
        r = requests.get(url).json()
        if len(r["results"]):
            text_sig = r["results"][-1]["text_signature"]
        else:
            text_sig = f"{method_sig}()"

        decoded_methods[method_sig] = text_sig
    else:
        text_sig = decoded_methods.get(method_sig)

    return text_sig


def decode_sequencer_batch(data):
    BATCH_CONTEXT_START_POS = 15
    BATCH_CONTEXT_SIZE = 16
    TX_DATA_HEADER_SIZE = 3

    def load_call_data(data, position, shift):

        sub_data = data[2 + 2 * position :]
        value = int(sub_data[: shift * 2], 16)

        return value

    def load_tx_data(data, position, length):
        def ECDSA_recover(transaction):
            tx = Transaction(
                transaction["nonce"],
                transaction["gas_price"],
                transaction["gas_limit"],
                b""
                if transaction["to_address"]
                == "0x0000000000000000000000000000000000000000"
                else transaction["to_address"],
                transaction["value"],
                bytes.fromhex(transaction["data"]),
                int(transaction["v"], 16) + 55,
                int(transaction["r"], 16),
                int(transaction["s"], 16),
            )

            tx_hash = "0x" + tx.hash.hex()
            from_address = "0x" + tx.sender.hex()

            return from_address, tx_hash

        sub_data = data[2 + 2 * position :][: length * 2]

        is_eip155 = int(sub_data[:2])
        r = sub_data[2 : 33 * 2]
        s = sub_data[33 * 2 : 65 * 2]
        v = sub_data[65 * 2 : 66 * 2]
        gas_limit = int(sub_data[66 * 2 : 69 * 2], 16)
        gas_price = int(sub_data[69 * 2 : 72 * 2], 16)
        nonce = int(sub_data[72 * 2 : 75 * 2], 16)
        to_address = "0x" + sub_data[75 * 2 : 95 * 2]
        data = sub_data[95 * 2 :]
        signature = decode_from_4byte("0x" + data[:8], decoded_methods)
        input_data = data[8:]

        transaction = dict(
            eip155=(is_eip155 == 0),
            r=r,
            s=s,
            v=v,
            gas_limit=gas_limit,
            gas_price=gas_price,
            nonce=nonce,
            to_address=to_address,
            value=0,
            data=data,
            signature=signature,
            input=input_data,
        )

        transaction["from_address"], transaction["tx_hash"] = ECDSA_recover(transaction)
        transaction.pop("data")

        return transaction

    decoded_methods = dict()

    data = "0x00000000" + data

    shouldStartAtElement = load_call_data(data, 4, 5)
    totalElementsToAppend = load_call_data(data, 9, 3)
    numContexts = load_call_data(data, 12, 3)
    numTransactions = 0

    batch = dict(
        shouldStartAtElement=shouldStartAtElement,
        totalElementsToAppend=totalElementsToAppend,
        numContexts=numContexts,
        contexts=[],
    )

    nextTransactionPtr = BATCH_CONTEXT_START_POS + BATCH_CONTEXT_SIZE * numContexts

    for i in range(numContexts):

        contextPtr = 15 + i * BATCH_CONTEXT_SIZE
        numSequencedTransactions = load_call_data(data, contextPtr, 3)
        numSubsequentQueueTransactions = load_call_data(data, contextPtr + 3, 3)
        ctxTimestamp = datetime.utcfromtimestamp(
            load_call_data(data, contextPtr + 6, 5)
        )
        ctxBlockNumber = load_call_data(data, contextPtr + 11, 5)

        context = dict(
            numSequencedTransactions=numSequencedTransactions,
            numSubsequentQueueTransactions=numSubsequentQueueTransactions,
            ctxTimestamp=ctxTimestamp,
            ctxBlockNumber=ctxBlockNumber,
            ctxSequencedTransactions=[],
        )

        for _ in range(numSequencedTransactions):
            txDataLength = load_call_data(data, nextTransactionPtr, 3)
            transactionData = load_tx_data(data, nextTransactionPtr + 3, txDataLength)
            context["ctxSequencedTransactions"].append(transactionData)
            numTransactions += 1
            nextTransactionPtr += TX_DATA_HEADER_SIZE + txDataLength

        batch["contexts"].append(context)

    batch["numTransactions"] = numTransactions

    return batch


def decode_ovm_message(data):
    target = gas_limit = data
    signature = input_data = None
    transaction = dict(
        eip155=Transaction,
        r=None,
        s=None,
        v=None,
        gas_limit=gas_limit,
        gas_price=0,
        nonce=0,
        to_address=target,
        value=0,
        data=data,
        signature=signature,
        input=input_data,
    )

    context = dict(
        ctxTimestamp=None, ctxBlockNumber=None, ctxSequencedTransactions=[transaction]
    )

    batch = dict(numContexts=1, contexts=[context])

    return batch



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional, Dict

from ethtx.models.decoded_model import DecodedCall
from ethtx.models.objects_model import Call, TransactionMetadata
from ethtx.utils.measurable import RecursionLimit
from .abc import ABISubmoduleAbc
from ..decoders.parameters import decode_function_parameters, decode_graffiti_parameters

RECURSION_LIMIT = 2000


class ABICallsDecoder(ABISubmoduleAbc):
    """Abi Calls Decoder."""

    def decode(
        self,
        call: Call,
        transaction: TransactionMetadata,
        delegations: Optional[Dict[str, set]] = None,
        token_proxies: Optional[Dict[str, dict]] = None,
        chain_id: Optional[str] = None,
    ) -> Optional[DecodedCall]:
        """Decode call with sub_calls."""

        if not call:
            return None

        indent = 0
        status = True
        call_id = ""

        decoded_root_call = self.decode_call(
            call,
            transaction,
            call_id,
            indent,
            status,
            delegations,
            token_proxies,
            chain_id,
        )

        with RecursionLimit(RECURSION_LIMIT):
            calls_tree = self._decode_nested_calls(
                decoded_root_call,
                transaction,
                call.subcalls,
                indent,
                status,
                delegations,
                token_proxies,
            )

        calls_tree = self._prune_delegates(calls_tree)

        return calls_tree

    def decode_call(
        self,
        call: Call,
        transaction: TransactionMetadata,
        call_id: str = "",
        indent: int = 0,
        status: bool = True,
        delegations: Dict[str, set] = None,
        token_proxies: Dict[str, dict] = None,
        chain_id: str = None,
    ) -> DecodedCall:
        """Decode single call."""
        if call.call_data:
            function_signature = call.call_data[:10]
        else:
            function_signature = None

        from_name = self._repository.get_address_label(
            chain_id, call.from_address, token_proxies
        )
        to_name = self._repository.get_address_label(
            chain_id, call.to_address, token_proxies
        )

        delegations = delegations or {}

        if call.call_type == "selfdestruct":
            function_name = call.call_type
            function_input, function_output = [], []
        elif call.call_type == "create2":
            # constructor_abi = self.repository.get_constructor_abi(call.chain_id, call.to_address)
            function_name = "new"
            function_input, function_output = [], []
        elif self._repository.check_is_contract(chain_id, call.to_address):

            function_abi = self._repository.get_function_abi(
                chain_id, call.to_address, function_signature
            )
            if not function_abi and call.to_address in delegations:
                # try to find signature in delegate-called contracts
                for delegate in delegations[call.to_address]:
                    function_abi = self._repository.get_function_abi(
                        chain_id, delegate, function_signature
                    )
                    if function_abi:
                        break

            function_name = function_abi.name if function_abi else function_signature
            function_input, function_output = decode_function_parameters(
                call.call_data, call.return_value, function_abi, call.status
            )
            if (
                not call.status
                and function_output
                and function_output[0].name == "Error"
            ):
                error_description = function_output.pop()
                call.error = f'Failed with "{error_description.value}"'
        else:
            function_name = "fallback"
            function_input = decode_graffiti_parameters(call.call_data)
            function_output = []

        return DecodedCall(
            chain_id=chain_id,
            tx_hash=transaction.tx_hash,
            call_id=call_id,
            call_type=call.call_type,
            from_address=call.from_address,
            from_name=from_name,
            to_address=call.to_address,
            to_name=to_name,
            value=call.call_value / 10 ** 18,
            function_signature=function_signature,
            function_name=function_name,
            arguments=function_input,
            outputs=function_output,
            gas_used=call.gas_used,
            error=call.error,
            status=status,
            indent=indent,
        )

    def _decode_nested_calls(
        self,
        call: DecodedCall,
        transaction: TransactionMetadata,
        sub_calls,
        indent,
        status,
        delegations,
        token_proxies,
    ) -> DecodedCall:
        """Decode nested calls. Call may have sub_calls, if they exist, it will recursively process them."""
        for i, sub_call in enumerate(sub_calls):
            status = status and call.status

            sub_call_id = (
                "_".join([call.call_id, str(i).zfill(4)]) if call.call_id else str(i)
            )
            decoded = self.decode_call(
                sub_call,
                transaction,
                sub_call_id,
                indent + 1,
                status,
                delegations,
                token_proxies,
            )
            call.subcalls.append(decoded)

            if sub_call.subcalls:
                self._decode_nested_calls(
                    decoded,
                    transaction,
                    sub_call.subcalls,
                    indent + 1,
                    status,
                    delegations,
                    token_proxies,
                )

        return call

    def _prune_delegates(self, call: DecodedCall) -> DecodedCall:

        while len(call.subcalls) == 1 and call.subcalls[0].call_type == "delegatecall":
            call = call.subcalls[0]

        for i, sub_call in enumerate(call.subcalls):
            call.subcalls[i] = self._prune_delegates(sub_call)

        return call
